/***************************************************************************
 * RT2x00 SourceForge Project - http://rt2x00.serialmonkey.com             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   Licensed under the GNU GPL                                            *
 *   Original code supplied under license from RaLink Inc, 2004.           *
 ***************************************************************************/

/***************************************************************************
 *	Module Name:	rtmp_main.c
 *
 *	Abstract: Main initialization routines
 *
 *	Revision History:
 *	Who		When		What
 *	--------	----------	-----------------------------
 *	Jan Lee		01-10-2005	modified
 *	idamlaj		04-10-2006	Apply patch by Ace17 (from forum)
 *
 ***************************************************************************/

#include "rt_config.h"

#define RT73_DBUG_PRINTK

extern CHAR *GetEncryptType(CHAR enc, CHAR encBitMode);
extern CHAR *GetAuthMode(CHAR auth, CHAR authBitMode);

ULONG RTDebugLevel;
#ifdef RT73_DBUG_PRINTK
int dbug_level = 0;
int dbug_pages = 64; // max. # of 4K pages of debug text
module_param(dbug_level, int, 0);
module_param(dbug_pages, int, 0);
#endif
int zero_copy = 1;
int dbug_count = 0;

int debug = RT_DEBUG_WARN;
int config_short_slot_time = 1;
module_param(debug, int, 0);
module_param(config_short_slot_time, int, 0);

MODULE_PARM_DESC(debug, "debug: 0-OFF 1-ERROR 2-WARN 3-INFO 4-TRACE 5-LOUD");
MODULE_AUTHOR("http://rt2x00.serialmonkey.com");
MODULE_DESCRIPTION("Ralink RT73 802.11abg WLAN Driver " DRIVER_VERSION " " DRIVER_RELDATE);
MODULE_LICENSE("GPL");


/* Kernel thread and vars, which handles packets that are completed. Only
 * packets that have a "complete" function are sent here. This way, the
 * completion is run out of kernel context, and doesn't block the rest of
 * the stack. */
static int mlme_kill;
static int RTUSBCmd_kill;

extern const struct iw_handler_def rt73_iw_handler_def;


/* module table */
struct usb_device_id rtusb_usb_id[] = RT73_USB_DEVICES;
INT const rtusb_usb_id_len =
    sizeof(rtusb_usb_id) / sizeof(struct usb_device_id);
MODULE_DEVICE_TABLE(usb, rtusb_usb_id);


#ifndef PF_NOFREEZE
#define PF_NOFREEZE  0
#endif


/**************************************************************************/
/**************************************************************************/
//tested for kernel 2.4 series
/**************************************************************************/
/**************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)


static void usb_rtusb_disconnect(struct usb_device *dev, void *ptr);
static void *usb_rtusb_probe(struct usb_device *dev, UINT interface,
                             const struct usb_device_id *id_table);

struct usb_driver rtusb_driver = {
  name:"rt73",
  probe:usb_rtusb_probe,
  disconnect:usb_rtusb_disconnect,
  id_table:rtusb_usb_id,
};
#else
/**************************************************************************/
/**************************************************************************/
//tested for kernel 2.6series
/**************************************************************************/
/**************************************************************************/
static int usb_rtusb_probe(struct usb_interface *intf,
                           const struct usb_device_id *id);

static void usb_rtusb_disconnect(struct usb_interface *intf);

struct usb_driver rtusb_driver = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
    .owner = THIS_MODULE,
#endif
    .name = "rt73",
    .probe = usb_rtusb_probe,
    .disconnect = usb_rtusb_disconnect,
    .id_table = rtusb_usb_id,
};


#endif


static struct net_device_stats *rt73_get_ether_stats(IN struct net_device
                                              *net_dev)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) net_dev->priv;

    DBGPRINT(RT_DEBUG_INFO, "rt73_get_ether_stats --->\n");

    pAd->stats.rx_packets = pAd->WlanCounters.ReceivedFragmentCount.vv.LowPart; // total packets received
    pAd->stats.tx_packets = pAd->WlanCounters.TransmittedFragmentCount.vv.LowPart;      // total packets transmitted

    pAd->stats.rx_bytes = pAd->RalinkCounters.ReceivedByteCount;        // total bytes received
    pAd->stats.tx_bytes = pAd->RalinkCounters.TransmittedByteCount;     // total bytes transmitted

    pAd->stats.rx_errors = pAd->Counters8023.RxErrors;  // bad packets received
    pAd->stats.tx_errors = pAd->Counters8023.TxErrors;  // packet transmit problems

    pAd->stats.rx_dropped = pAd->Counters8023.RxNoBuffer;       // no space in linux buffers
    pAd->stats.tx_dropped = pAd->WlanCounters.FailedCount.vv.LowPart;   // no space available in linux

    pAd->stats.multicast = pAd->WlanCounters.MulticastReceivedFrameCount.vv.LowPart;    // multicast packets received
    pAd->stats.collisions = pAd->Counters8023.OneCollision + pAd->Counters8023.MoreCollisions;  // Collision packets

    pAd->stats.rx_length_errors = 0;
    pAd->stats.rx_over_errors = pAd->Counters8023.RxNoBuffer;   // receiver ring buff overflow
    pAd->stats.rx_crc_errors = 0;       //pAd->WlanCounters.FCSErrorCount;         // recved pkt with crc error
    pAd->stats.rx_frame_errors = pAd->Counters8023.RcvAlignmentErrors;  // recv'd frame alignment error
    pAd->stats.rx_fifo_errors = pAd->Counters8023.RxNoBuffer;   // recv'r fifo overrun
    pAd->stats.rx_missed_errors = 0;    // receiver missed packet

    // detailed tx_errors
    pAd->stats.tx_aborted_errors = 0;
    pAd->stats.tx_carrier_errors = 0;
    pAd->stats.tx_fifo_errors = 0;
    pAd->stats.tx_heartbeat_errors = 0;
    pAd->stats.tx_window_errors = 0;

    // for cslip etc
    pAd->stats.rx_compressed = 0;
    pAd->stats.tx_compressed = 0;

    return &pAd->stats;
}

#if WIRELESS_EXT >= 12
/*
	========================================================================

	Routine Description:
		get wireless statistics

	Arguments:
		net_dev                     Pointer to net_device

	Return Value:
		struct iw_statistics

	Note:
		This function will be called when query /proc

	========================================================================
*/
long rt_abs(long arg)
{
    return (arg < 0) ? -arg : arg;
}
struct iw_statistics *rt73_get_wireless_stats(IN struct net_device *net_dev)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) net_dev->priv;

    DBUG_HERE(4);

    // TODO: All elements are zero before be implemented

    pAd->iw_stats.status = 0;   // Status - device dependent for now

    pAd->iw_stats.qual.qual = pAd->Mlme.ChannelQuality; // link quality (%retries, SNR, %missed beacons or better...)
#ifdef RTMP_EMBEDDED
    pAd->iw_stats.qual.level = rt_abs(pAd->PortCfg.LastRssi);   // signal level (dBm)
#else
    pAd->iw_stats.qual.level = abs(pAd->PortCfg.LastRssi);      // signal level (dBm)
#endif
    pAd->iw_stats.qual.level += 256 - pAd->BbpRssiToDbmDelta;

    pAd->iw_stats.qual.noise = (pAd->BbpWriteLatch[17] > pAd->BbpTuning.R17UpperBoundG) ? pAd->BbpTuning.R17UpperBoundG : ((ULONG) pAd->BbpWriteLatch[17]);     // noise level (dBm)
    pAd->iw_stats.qual.noise += 256 - 143;
    pAd->iw_stats.qual.updated = 1;     // Flags to know if updated

    pAd->iw_stats.discard.nwid = 0;     // Rx : Wrong nwid/essid
    pAd->iw_stats.miss.beacon = 0;      // Missed beacons/superframe

    // pAd->iw_stats.discard.code, discard.fragment, discard.retries, discard.misc has counted in other place

    return &pAd->iw_stats;
}
#endif

static VOID RTUSBHalt(IN PRTMP_ADAPTER pAd, IN BOOLEAN IsFree)
{
    MLME_DISASSOC_REQ_STRUCT DisReq;
    MLME_QUEUE_ELEM MsgElem;
    INT i;

    DBUG_HERE(4);

    //
    // before set flag fRTMP_ADAPTER_HALT_IN_PROGRESS,
    // we should send a disassoc frame to our AP.
    //
    if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)) {
        if (INFRA_ON(pAd)) {
            COPY_MAC_ADDR(DisReq.Addr, pAd->PortCfg.Bssid);
            DisReq.Reason = REASON_DISASSOC_STA_LEAVING;

            MsgElem.Machine = ASSOC_STATE_MACHINE;
            MsgElem.MsgType = MT2_MLME_DISASSOC_REQ;
            MsgElem.MsgLen = sizeof(MLME_DISASSOC_REQ_STRUCT);
            memcpy(MsgElem.Msg, &DisReq, sizeof(MLME_DISASSOC_REQ_STRUCT));

            MlmeDisassocReqAction(pAd, &MsgElem);
            RTMPusecDelay(1000);
        }
        //
        // Patch to fully turn off BBP, need to send a fake NULL frame.
        //
        RTUSBWriteMACRegister(pAd, MAC_CSR10, 0x0018);
        for (i = 0; i < 10; i++) {
            RTMPSendNullFrame(pAd, RATE_6);
            RTMPusecDelay(500); // GRH 2/19/2008 -- changed from one 1000usec
            RTMPusecDelay(500); // delay to two 500usec delays, refer to 
                                // comments in RTMPusecDelay() for details
        }

        // disable BEACON generation and other BEACON related hardware timers
        AsicDisableSync(pAd);
        RTMPSetLED(pAd, LED_HALT);

    }

    RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS);

    RTUSBCleanUpMLMEWaitQueue(pAd);
    RTUSBCleanUpMLMEBulkOutQueue(pAd);

    RTMPCancelTimer(&pAd->PortCfg.QuickResponeForRateUpTimer);
    RTMPCancelTimer(&pAd->RxAnt.RxAntDiversityTimer);

    // Free MLME stuff
    MlmeHalt(pAd);

    // Sleep 50 milliseconds so pending io might finish normally
    RTMPusecDelay(50000);

    // We want to wait until all pending receives and sends to the
    // device object. We cancel any
    // irps. Wait until sends and receives have stopped.
    //
    RTUSBCancelPendingIRPs(pAd);

    // Free the entire adapter object
    ReleaseAdapter(pAd, IsFree, FALSE);
    OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED);

    // reset mlme & command thread
    pAd->MLMEThr_pid = -1;
    pAd->RTUSBCmdThr_pid = -1;

    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS);
}

char* CMDString( int command )
{
    switch (command) {
    case RT_OID_CHECK_GPIO: return "RT_OID_CHECK_GPIO";
    case RT_OID_PERIODIC_EXECUT: return "RT_OID_PERIODIC_EXECUT";
    case OID_802_11_BSSID_LIST_SCAN: return "OID_802_11_BSSID_LIST_SCAN";
    case RT_OID_802_11_BSSID: return "RT_OID_802_11_BSSID";
    case OID_802_11_SSID: return "OID_802_11_SSID";
    case OID_802_11_DISASSOCIATE: return "OID_802_11_DISASSOCIATE";
    case OID_802_11_RX_ANTENNA_SELECTED: return "OID_802_11_RX_ANTENNA_SELECTED";
    case OID_802_11_TX_ANTENNA_SELECTED: return "OID_802_11_TX_ANTENNA_SELECTED";
    case RT_OID_MULTI_READ_MAC: return "RT_OID_MULTI_READ_MAC";
    case RT_OID_MULTI_WRITE_MAC: return "RT_OID_MULTI_WRITE_MAC";
    case RT_OID_USB_VENDOR_EEPROM_WRITE: return "RT_OID_USB_VENDOR_EEPROM_WRITE";
    case RT_OID_USB_VENDOR_ENTER_TESTMODE: return "RT_OID_USB_VENDOR_ENTER_TESTMODE";
    case RT_OID_USB_VENDOR_EXIT_TESTMODE: return "RT_OID_USB_VENDOR_EXIT_TESTMODE";
    case RT_OID_USB_RESET_BULK_OUT: return "RT_OID_USB_RESET_BULK_OUT";
    case RT_OID_USB_RESET_BULK_IN: return "RT_OID_USB_RESET_BULK_IN";
    case RT_OID_802_11_STA_CONFIG: return "RT_OID_802_11_STA_CONFIG";
    case RT_OID_SET_PSM_BIT_SAVE: return "RT_OID_SET_PSM_BIT_SAVE";
    case RT_OID_SET_RADIO: return "RT_OID_SET_RADIO";
    case RT_OID_RESET_FROM_ERROR: return "RT_OID_RESET_FROM_ERROR";
    case RT_OID_RESET_FROM_NDIS: return "";
    case RT_OID_LINK_DOWN: return "RT_OID_LINK_DOWN";
    case RT_OID_VENDOR_WRITE_BBP: return "RT_OID_VENDOR_WRITE_BBP";
    case RT_OID_VENDOR_READ_BBP: return "RT_OID_VENDOR_READ_BBP";
    case RT_OID_VENDOR_WRITE_RF: return "RT_OID_VENDOR_WRITE_RF";
    case RT_OID_802_11_RESET_COUNTERS: return "RT_OID_802_11_RESET_COUNTERS";
    case RT_OID_USB_VENDOR_RESET: return "RT_OID_USB_VENDOR_RESET";
    case RT_OID_USB_VENDOR_UNPLUG: return "RT_OID_USB_VENDOR_UNPLUG";
    case RT_OID_VENDOR_FLIP_IQ: return "RT_OID_VENDOR_FLIP_IQ";
    case RT_OID_UPDATE_TX_RATE: return "RT_OID_UPDATE_TX_RATE";
    case RT_OID_802_11_PREAMBLE: return "RT_OID_802_11_PREAMBLE";
    case OID_802_11_NETWORK_TYPE_IN_USE: return "OID_802_11_NETWORK_TYPE_IN_USE";
    case RT_OID_802_11_PHY_MODE: return "RT_OID_802_11_PHY_MODE";
    case OID_802_11_ADD_WEP: return "OID_802_11_ADD_WEP";
    case OID_802_11_REMOVE_WEP: return "OID_802_11_REMOVE_WEP";
    case OID_802_11_ADD_KEY_WEP: return "OID_802_11_ADD_KEY_WEP";
    case OID_802_11_ADD_KEY: return "OID_802_11_ADD_KEY";
    case OID_802_11_REMOVE_KEY: return "OID_802_11_REMOVE_KEY";
    case OID_802_11_POWER_MODE: return "OID_802_11_POWER_MODE";
    case RT_PERFORM_SOFT_DIVERSITY: return "RT_PERFORM_SOFT_DIVERSITY";
    case RT_OID_FORCE_WAKE_UP: return "RT_OID_FORCE_WAKE_UP";
    case RT_OID_SET_PSM_BIT_ACTIVE: return "RT_OID_SET_PSM_BIT_ACTIVE";
    case MT2_MLME_ROAMING_REQ: return "MT2_MLME_ROAMING_REQ";

    default:
        {
        static char unknown[32];
        sprintf(unknown, "0x%x", command );
        return unknown;
        }
    }
}


static VOID CMDHandler(IN PRTMP_ADAPTER pAd)
{
    PCmdQElmt cmdqelmt;
    PUCHAR pData;
    NDIS_STATUS NdisStatus = NDIS_STATUS_SUCCESS;
    unsigned long IrqFlags;
    ULONG Now;

    while (pAd->CmdQ.size > 0) {
        NdisStatus = NDIS_STATUS_SUCCESS;
        NdisAcquireSpinLock(&pAd->CmdQLock, IrqFlags);
        RTUSBDequeueCmd(&pAd->CmdQ, &cmdqelmt);
        NdisReleaseSpinLock(&pAd->CmdQLock, IrqFlags);
        if (cmdqelmt == NULL)
            break;
        pData = cmdqelmt->buffer;

        if (cmdqelmt->command == RT_OID_PERIODIC_EXECUT)
            DBUG(9, "command=%s...\n", CMDString(cmdqelmt->command));
        else
            DBUG(3, "command=%s...\n", CMDString(cmdqelmt->command));

        switch (cmdqelmt->command) {
        case RT_OID_CHECK_GPIO:
            {
                ULONG data;
                // Read GPIO pin7 as Hardware controlled radio state
                RTUSBReadMACRegister(pAd, MAC_CSR13, &data);
                if (data & 0x80) {
                    pAd->PortCfg.bHwRadio = TRUE;
                } else {
                    pAd->PortCfg.bHwRadio = FALSE;
                }
                if (pAd->PortCfg.bRadio !=
                    (pAd->PortCfg.bHwRadio && pAd->PortCfg.bSwRadio)) {
                    pAd->PortCfg.bRadio = (pAd->PortCfg.bHwRadio &&
                                           pAd->PortCfg.bSwRadio);
                    if (pAd->PortCfg.bRadio == TRUE) {
                        ERR("RT_OID_CHECK_GPIO: calling MlmeRadioOn()...\n");
                        MlmeRadioOn(pAd);
                        // Update extra information
                        pAd->ExtraInfo = EXTRA_INFO_CLEAR;
                    } else {
                        MlmeRadioOff(pAd);
                        // Update extra information
                        pAd->ExtraInfo = HW_RADIO_OFF;
                    }
                }
            }
            break;

        case RT_OID_PERIODIC_EXECUT:
            STAMlmePeriodicExec(pAd);
            break;

        case OID_802_11_BSSID_LIST_SCAN:
            {
                if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE) {
                    MlmeEnqueue(pAd,
                                MLME_CNTL_STATE_MACHINE,
                                RT_CMD_RESET_MLME, 0, NULL);

                }

                Now = jiffies;
                pAd->MlmeAux.CurrReqIsFromNdis = FALSE;
                // Reset Missed scan number
                pAd->PortCfg.ScanCnt = 0;
                pAd->PortCfg.LastScanTime = Now;
                MlmeEnqueue(pAd,
                            MLME_CNTL_STATE_MACHINE,
                            OID_802_11_BSSID_LIST_SCAN, 0, NULL);
                RTUSBMlmeUp(pAd);
            }
            break;

        case RT_OID_802_11_BSSID:
            {
                if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE) {
                    MlmeEnqueue(pAd,
                                MLME_CNTL_STATE_MACHINE,
                                RT_CMD_RESET_MLME, 0, NULL);
                }

                pAd->MlmeAux.CurrReqIsFromNdis = FALSE;

                // Reset allowed scan retries
                pAd->PortCfg.ScanCnt = 0;

                MlmeEnqueue(pAd,
                            MLME_CNTL_STATE_MACHINE,
                            OID_802_11_BSSID,
                            cmdqelmt->bufferlength, cmdqelmt->buffer);
                RTUSBMlmeUp(pAd);
            }
            break;

        case OID_802_11_SSID:
            {
                if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE) {
                    MlmeEnqueue(pAd,
                                MLME_CNTL_STATE_MACHINE,
                                RT_CMD_RESET_MLME, 0, NULL);
                }

                pAd->MlmeAux.CurrReqIsFromNdis = FALSE;

                // Reset allowed scan retries
                pAd->PortCfg.ScanCnt = 0;
                pAd->bConfigChanged = TRUE;

                MlmeEnqueue(pAd,
                            MLME_CNTL_STATE_MACHINE,
                            OID_802_11_SSID,
                            cmdqelmt->bufferlength, pData);
                RTUSBMlmeUp(pAd);
            }
            break;

        case OID_802_11_DISASSOCIATE:
            {
                if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE) {
                    MlmeEnqueue(pAd,
                                MLME_CNTL_STATE_MACHINE,
                                RT_CMD_RESET_MLME, 0, NULL);
                }
                // Set to immediately send the media disconnect event
                pAd->MlmeAux.CurrReqIsFromNdis = TRUE;

                MlmeEnqueue(pAd,
                            MLME_CNTL_STATE_MACHINE,
                            OID_802_11_DISASSOCIATE, 0, NULL);
                RTUSBMlmeUp(pAd);
            }
            break;

        case OID_802_11_RX_ANTENNA_SELECTED:
            {
                NDIS_802_11_ANTENNA Antenna =
                    *(NDIS_802_11_ANTENNA *) pData;

                if (Antenna == 0)
                    pAd->Antenna.field.RxDefaultAntenna = 1;    // ant-A
                else if (Antenna == 1)
                    pAd->Antenna.field.RxDefaultAntenna = 2;    // ant-B
                else
                    pAd->Antenna.field.RxDefaultAntenna = 0;    // diversity

                pAd->PortCfg.BandState = UNKNOWN_BAND;
                AsicAntennaSelect(pAd, pAd->LatchRfRegs.Channel);
                DBUG(4,"CMDHandler::OID_802_11_RX_ANTENNA_SELECTED (=%d)\n",
                         Antenna);
            }
            break;

        case OID_802_11_TX_ANTENNA_SELECTED:
            {
                NDIS_802_11_ANTENNA Antenna =
                    *(NDIS_802_11_ANTENNA *) pData;

                if (Antenna == 0)
                    pAd->Antenna.field.TxDefaultAntenna = 1;    // ant-A
                else if (Antenna == 1)
                    pAd->Antenna.field.TxDefaultAntenna = 2;    // ant-B
                else
                    pAd->Antenna.field.TxDefaultAntenna = 0;    // diversity

                pAd->PortCfg.BandState = UNKNOWN_BAND;
                AsicAntennaSelect(pAd, pAd->LatchRfRegs.Channel);
                DBUG(4, "CMDHandler::OID_802_11_TX_ANTENNA_SELECTED (=%d)\n",
                         Antenna);
            }
            break;

        case RT_OID_MULTI_READ_MAC:
            {
                USHORT Offset = *((PUSHORT) pData);
                USHORT Length = *((PUSHORT) (pData + 2));
                RTUSBMultiRead(pAd, Offset, pData + 4, Length);
            }
            break;

        case RT_OID_MULTI_WRITE_MAC:
            {
                USHORT Offset = *((PUSHORT) pData);
                USHORT Length = *((PUSHORT) (pData + 2));
                RTUSBMultiWrite(pAd, Offset, pData + 4, Length);
            }
            break;

        case RT_OID_USB_VENDOR_EEPROM_READ:
            {
                USHORT Offset = *((PUSHORT) pData);
                USHORT Length = *((PUSHORT) (pData + 2));
                RTUSBReadEEPROM(pAd, Offset, pData + 4, Length);
            }
            break;

        case RT_OID_USB_VENDOR_EEPROM_WRITE:
            {
                USHORT Offset = *((PUSHORT) pData);
                USHORT Residual = *((PUSHORT) (pData + 2));
                pData += 4;
                while (Residual > 62) {
                    RTUSBWriteEEPROM(pAd, Offset, pData, 62);
                    Offset += 62;
                    Residual -= 62;
                    pData += 62;
                }
                RTUSBWriteEEPROM(pAd, Offset, pData, Residual);
            }
            break;

        case RT_OID_USB_VENDOR_ENTER_TESTMODE:
            RTUSB_VendorRequest(pAd,
                                0,
                                DEVICE_VENDOR_REQUEST_OUT,
                                0x1, 0x4, 0x1, NULL, 0);
            break;

        case RT_OID_USB_VENDOR_EXIT_TESTMODE:
            RTUSB_VendorRequest(pAd,
                                0,
                                DEVICE_VENDOR_REQUEST_OUT,
                                0x1, 0x4, 0x0, NULL, 0);
            break;
        case RT_OID_USB_RESET_BULK_OUT:
            {
                INT Index;

                DBGPRINT_RAW(RT_DEBUG_TRACE, "RT_OID_USB_RESET_BULK_OUT\n");

                RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS);

                RTUSBRejectPendingPackets(pAd); //reject all NDIS packets waiting in TX queue
                RTUSBCancelPendingBulkOutIRP(pAd);
                RTUSBCleanUpDataBulkOutQueue(pAd);

                NICInitializeAsic(pAd);
                ReleaseAdapter(pAd, FALSE, TRUE);       // unlink urb releated tx context
                NICInitTransmit(pAd);

                RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS);

                if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET)) {
                    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
                }

                if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) {
                    for (Index = 0; Index < 4; Index++) {
                        if (!skb_queue_empty(&pAd->SendTxWaitQueue[Index])) {
                            RTMPDeQueuePacket(pAd, Index);
                        }
                    }
                    RTUSBKickBulkOut(pAd);
                }
            }

            break;

        case RT_OID_USB_RESET_BULK_IN:
            {
                int i;
                ERR("!!!!!RT_OID_USB_RESET_BULK_IN\n");
                RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS);
                NICInitializeAsic(pAd);
                //RTUSBWriteMACRegister(pAd, TXRX_CSR0, 0x025eb032); // ??
                for (i = 0; i < RX_RING_SIZE; i++) {
                    PRX_CONTEXT pRxContext = &(pAd->RxContext[i]);

                    if (pRxContext->pUrb != NULL) {
                        DBGPRINT( RT_DEBUG_TRACE, "usb_kill_urb(urb=%p, buff=%p)\n",
                                  pRxContext->pUrb, pRxContext->TransferBuffer );
                        RTUSB_UNLINK_URB(pRxContext->pUrb);
                        DBGPRINT( RT_DEBUG_TRACE, "usb_free_urb(%p)\n", pRxContext->pUrb );
                        usb_free_urb(pRxContext->pUrb);
                        pRxContext->pUrb = NULL;
                    }
                    #ifdef USE_RX_SKB
                    if (pRxContext->skb != NULL) {
                        RTUSBFreeSkbBuffer(pRxContext->skb);
                        pRxContext->skb = NULL;
                        pRxContext->TransferBuffer = NULL;
                    }
                    #else
                    if (pRxContext->TransferBuffer != NULL) {
                        kfree(pRxContext->TransferBuffer);
                        pRxContext->TransferBuffer = NULL;
                    }
                    #endif
                }
                NICInitRecv(pAd);
                RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS);
                if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET)) {
                    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);
                }

                if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) {
                    RTUSBBulkReceive(pAd);
                    RTUSBWriteMACRegister(pAd, TXRX_CSR0, 0x0276b032);  // enable RX of MAC block
                }
            }
            break;

        case RT_OID_802_11_STA_CONFIG:
            {
                RT_802_11_STA_CONFIG *pStaConfig =
                    (RT_802_11_STA_CONFIG *) pData;
                if (pStaConfig->EnableTxBurst !=
                    pAd->PortCfg.bEnableTxBurst) {
                    pAd->PortCfg.bEnableTxBurst =
                        (pStaConfig->EnableTxBurst == 1);
                    //Currently Tx burst mode is only implemented in infrastructure mode.
                    if (INFRA_ON(pAd)) {
                        if (pAd->PortCfg.bEnableTxBurst) {
                            //Extend slot time if any encryption method is used to give ASIC more time to do encryption/decryption during Tx burst mode.
                            if (pAd->PortCfg.WepStatus !=
                                Ndis802_11EncryptionDisabled) {
                                // Nemo  RT2573USBWriteMACRegister_old(pAd, MAC_CSR10, 0x20);
                            }
                            //Set CWmin/CWmax to 0.
                            // Nemo 2004    RT2573USBWriteMACRegister_old(pAd, MAC_CSR22, 0x100);
                        } else {
                            if (pAd->PortCfg.WepStatus !=
                                Ndis802_11EncryptionDisabled)
                                AsicSetSlotTime(pAd,
                                                (BOOLEAN) pAd->PortCfg.
                                                UseShortSlotTime);
                            // Nemo 2004    RT2573USBWriteMACRegister_old(pAd, MAC_CSR22, 0x53);
                        }
                    }
                }
                //pAd->PortCfg.EnableTurboRate = pStaConfig->EnableTurboRate;
                pAd->PortCfg.UseBGProtection = pStaConfig->UseBGProtection;
                //pAd->PortCfg.UseShortSlotTime = pStaConfig->UseShortSlotTime;
                pAd->PortCfg.UseShortSlotTime = 1;      // 2003-10-30 always SHORT SLOT capable
                if (pAd->PortCfg.AdhocMode != pStaConfig->AdhocMode) {
                    // allow dynamic change of "USE OFDM rate or not" in ADHOC mode
                    // if setting changed, need to reset current TX rate as well as BEACON frame format
                    pAd->PortCfg.AdhocMode = pStaConfig->AdhocMode;
                    if (pAd->PortCfg.BssType == BSS_ADHOC) {
                        MlmeUpdateTxRates(pAd, FALSE);
                        MakeIbssBeacon(pAd);
                        AsicEnableIbssSync(pAd);
                    }
                }
                DBUG(4, "CmdThread::RT_OID_802_11_SET_STA_CONFIG (Burst=%d,BGprot=%d,ShortSlot=%d,Adhoc=%d,Protection=%d\n",
                         pStaConfig->EnableTxBurst,
                         pStaConfig->UseBGProtection,
                         pStaConfig->UseShortSlotTime,
                         pStaConfig->AdhocMode,
                         pAd->PortCfg.UseBGProtection);
            }
            break;

        case RT_OID_SET_PSM_BIT_SAVE:
            MlmeSetPsmBit(pAd, PWR_SAVE);
            RTMPSendNullFrame(pAd, pAd->PortCfg.TxRate);
            break;

        case RT_OID_SET_RADIO:
            if (pAd->PortCfg.bRadio == TRUE) {
                ERR("RT_OID_SET_RADIO: calling MlmeRadioOn()...\n");
                MlmeRadioOn(pAd);
                // Update extra information
                pAd->ExtraInfo = EXTRA_INFO_CLEAR;
            } else {
                MlmeRadioOff(pAd);
                // Update extra information
                pAd->ExtraInfo = SW_RADIO_OFF;
            }
            break;

        case RT_OID_RESET_FROM_ERROR:
        case RT_OID_RESET_FROM_NDIS:
            {
                UINT i = 0;

                RTUSBRejectPendingPackets(pAd); //reject all NDIS packets waiting in TX queue
                RTUSBCleanUpDataBulkOutQueue(pAd);
                MlmeSuspend(pAd, FALSE);

                //Add code to access necessary registers here.
                //disable Rx
                RTUSBWriteMACRegister(pAd, TXRX_CSR2, 1);
                //Ask our device to complete any pending bulk in IRP.
                while ((atomic_read(&pAd->PendingRx) > 0) ||
                       (pAd->BulkOutPending[0] == TRUE) ||
                       (pAd->BulkOutPending[1] == TRUE) ||
                       (pAd->BulkOutPending[2] == TRUE) ||
                       (pAd->BulkOutPending[3] == TRUE))
                {
                    if (atomic_read(&pAd->PendingRx) > 0) {
                        DBGPRINT_RAW(RT_DEBUG_TRACE,
                                     "BulkIn IRP Pending!!!\n");
                        RTUSB_VendorRequest(pAd, 0,
                                            DEVICE_VENDOR_REQUEST_OUT,
                                            0x0C, 0x0, 0x0, NULL, 0);
                    }

                    if ((pAd->BulkOutPending[0] == TRUE) ||
                        (pAd->BulkOutPending[1] == TRUE) ||
                        (pAd->BulkOutPending[2] == TRUE) ||
                        (pAd->BulkOutPending[3] == TRUE)) {
                        DBGPRINT_RAW(RT_DEBUG_TRACE,
                                     "BulkOut IRP Pending!!!\n");
                        if (i == 0) {
                            i++;
                            RTUSBCancelPendingBulkOutIRP(pAd);
                        }
                    }

                    RTMPusecDelay(500000);
                }

                #if 0
                // this resets all of the RT73 registers, but then after this
                // routine is finished, the system expects everything to be
                // fine -- correct keys are set, etc.  
                // 
                // this routine also sets some pAd state information -- so
                // either we don't do this reset or we have to redo everything
                // it undoes -- GRH 2/20/2008
                // 
                NICResetFromError(pAd);
                #endif

                if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HARDWARE_ERROR)) {
                    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_HARDWARE_ERROR);
                }
                if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET)) {
                    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);
                }
                if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET)) {
                    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
                }

                RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS);

                if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF        |
                                         fRTMP_ADAPTER_HALT_IN_PROGRESS |
                                         fRTMP_ADAPTER_NIC_NOT_EXIST))
                {
                    MlmeResume(pAd);
                    RTUSBBulkReceive(pAd);
                    RTUSBWriteMACRegister(pAd, TXRX_CSR2, 0x8a8b8c8d/*0x7e*/);

                    #if 0
                    // NICResetFromError() is called above, there is a bunch of
                    // stuff that we have to add here -- this is a start.
                    // So, the question is what is this reset routine trying
                    // to recover from?  If the hardware is not programmed 
                    // correctly we need all of this code.  If it is some sort
                    // of system issue in the driver maybe we don't.  
                    // 
                    // Currently this code is only called by in the 
                    // RT_OID_RESET_FROM_ERROR case when trying to recover 
                    // from the condition where the MLME TXQ is full -- GRH 2/08
                    // 
                    RTUSBWriteMACRegister(pAd, TXRX_CSR0, 0x025eb032);      // enable RX of MAC block, Station not drop control frame
                    AsicSetBssid(pAd, pAd->PortCfg.Bssid);
                    AsicSetSlotTime(pAd, (BOOLEAN) pAd->PortCfg.UseShortSlotTime);
                    AsicSetEdcaParm(pAd, &pAd->PortCfg.APEdcaParm);
                    MlmeUpdateTxRates(pAd, TRUE);
                    #endif
                }
            }
            break;

        case RT_OID_LINK_DOWN:
            DBGPRINT_RAW(RT_DEBUG_TRACE, "LinkDown(RT_OID_LINK_DOWN)\n");
            LinkDown(pAd, TRUE);
            break;

        case RT_OID_VENDOR_WRITE_BBP:
            {
                UCHAR Offset, Value;
                Offset = *((PUCHAR) pData);
                Value = *((PUCHAR) (pData + 1));
                DBGPRINT_RAW(RT_DEBUG_INFO,
                             "offset = 0x%02x	value = 0x%02x\n", Offset,
                             Value);
                RTUSBWriteBBPRegister(pAd, Offset, Value);
            }
            break;

        case RT_OID_VENDOR_READ_BBP:
            {
                UCHAR Offset = *((PUCHAR) pData);
                PUCHAR pValue = (PUCHAR) (pData + 1);

                DBGPRINT_RAW(RT_DEBUG_INFO, "offset = 0x%02x\n", Offset);
                RTUSBReadBBPRegister(pAd, Offset, pValue);
                DBGPRINT_RAW(RT_DEBUG_INFO, "value = 0x%02x\n", *pValue);
            }
            break;

        case RT_OID_VENDOR_WRITE_RF:
            {
                ULONG Value = *((PULONG) pData);

                DBGPRINT_RAW(RT_DEBUG_INFO, "value = 0x%08x\n", Value);
                RTUSBWriteRFRegister(pAd, Value);
            }
            break;

        case RT_OID_802_11_RESET_COUNTERS:
            {
                UCHAR Value[24];

                RTUSBMultiRead(pAd, STA_CSR0, Value, 24);
            }
            break;

        case RT_OID_USB_VENDOR_RESET:
            RTUSB_VendorRequest(pAd,
                                0,
                                DEVICE_VENDOR_REQUEST_OUT,
                                1, 1, 0, NULL, 0);
            break;

        case RT_OID_USB_VENDOR_UNPLUG:
            RTUSB_VendorRequest(pAd,
                                0,
                                DEVICE_VENDOR_REQUEST_OUT,
                                1, 2, 0, NULL, 0);
            break;

        case RT_OID_VENDOR_FLIP_IQ:
            {
                ULONG Value1, Value2;
                RTUSBReadMACRegister(pAd, PHY_CSR5, &Value1);
                RTUSBReadMACRegister(pAd, PHY_CSR6, &Value2);
                if (*pData == 1) {
                    DBGPRINT_RAW(RT_DEBUG_INFO, "I/Q Flip\n");
                    Value1 = Value1 | 0x0004;
                    Value2 = Value2 | 0x0004;
                } else {
                    DBGPRINT_RAW(RT_DEBUG_INFO, "I/Q Not Flip\n");
                    Value1 = Value1 & 0xFFFB;
                    Value2 = Value2 & 0xFFFB;
                }
                RTUSBWriteMACRegister(pAd, PHY_CSR5, Value1);
                RTUSBWriteMACRegister(pAd, PHY_CSR6, Value2);
            }
            break;

        case RT_OID_UPDATE_TX_RATE:
            MlmeUpdateTxRates(pAd, FALSE);
            if (ADHOC_ON(pAd))
                MakeIbssBeacon(pAd);
            break;

        case RT_OID_802_11_PREAMBLE:
            {
                ULONG Preamble = *((PULONG) (cmdqelmt->buffer));
                if (Preamble == Rt802_11PreambleShort) {
                    pAd->PortCfg.TxPreamble = Preamble;
                    MlmeSetTxPreamble(pAd, Rt802_11PreambleShort);
                } else if ((Preamble == Rt802_11PreambleLong) ||
                           (Preamble == Rt802_11PreambleAuto)) {
                    // if user wants AUTO, initialize to LONG here, then change according to AP's
                    // capability upon association.
                    pAd->PortCfg.TxPreamble = Preamble;
                    MlmeSetTxPreamble(pAd, Rt802_11PreambleLong);
                } else
                    NdisStatus = NDIS_STATUS_FAILURE;
                DBGPRINT(RT_DEBUG_TRACE,
                         "CMDHandler::RT_OID_802_11_SET_PREAMBLE (=%d)\n",
                         Preamble);
            }
            break;
        case OID_802_11_NETWORK_TYPE_IN_USE:
            {
                NDIS_802_11_NETWORK_TYPE NetType =
                    *(PNDIS_802_11_NETWORK_TYPE) (cmdqelmt->buffer);
                if (NetType == Ndis802_11DS)
                    RTMPSetPhyMode(pAd, PHY_11B);
                else if (NetType == Ndis802_11OFDM24)
                    RTMPSetPhyMode(pAd, PHY_11BG_MIXED);
                else if (NetType == Ndis802_11OFDM5)
                    RTMPSetPhyMode(pAd, PHY_11A);
                else
                    NdisStatus = NDIS_STATUS_FAILURE;
                DBGPRINT(RT_DEBUG_TRACE,
                         "CMDHandler::OID_802_11_NETWORK_TYPE_IN_USE (=%d)\n",
                         NetType);
            }
            break;
        case RT_OID_802_11_PHY_MODE:
            {
                ULONG phymode = *(ULONG *) (cmdqelmt->buffer);
                RTMPSetPhyMode(pAd, phymode);
                INFO("RT_OID_802_11_SET_PHY_MODE (phymode=%ld)\n", phymode);
            }
            break;

        case OID_802_11_ADD_WEP:
            {
                ULONG KeyIdx;
                PNDIS_802_11_WEP pWepKey;

                DBUG(4,"CMDHandler::OID_802_11_ADD_WEP  \n");

                pWepKey = (PNDIS_802_11_WEP) pData;
                KeyIdx = pWepKey->KeyIndex & 0x0fffffff;

                // it is a shared key
                if ((KeyIdx >= 4) ||
                    ((pWepKey->KeyLength != 5) &&
                     (pWepKey->KeyLength != 13))) {
                    NdisStatus = NDIS_STATUS_FAILURE;
                    ERR("OID_802_11_ADD_WEP: KeyIdx %ld, KeyLength %ld, INVALID_DATA!!\n",
                        KeyIdx, pWepKey->KeyLength);
                } else {
                    UCHAR CipherAlg;
                    pAd->SharedKey[KeyIdx].KeyLen =
                        (UCHAR) pWepKey->KeyLength;
                    memcpy(pAd->SharedKey[KeyIdx].Key,
                           &pWepKey->KeyMaterial, pWepKey->KeyLength);
                    CipherAlg =
                        (pAd->SharedKey[KeyIdx].KeyLen ==
                         5) ? CIPHER_WEP64 : CIPHER_WEP128;
                    pAd->SharedKey[KeyIdx].CipherAlg = CipherAlg;
                    if (pWepKey->KeyIndex & 0x80000000) {
                        // Default key for tx (shared key)
                        pAd->PortCfg.DefaultKeyId = (UCHAR) KeyIdx;
                        DBUG(4, "GOTCHA: DefaultKeyId = %d\n", pAd->PortCfg.DefaultKeyId );
                    }
                    DBUG(4, "CMDHandler::OID_802_11_ADD_WEP (KeyIdx=%d, Len=%d-byte)\n",
                         KeyIdx, pWepKey->KeyLength);
                    AsicAddSharedKeyEntry(pAd, 0, (UCHAR) KeyIdx,
                                          CipherAlg, pWepKey->KeyMaterial,
                                          NULL, NULL);
                }
            }
            break;

        case OID_802_11_REMOVE_WEP:
            {
                ULONG KeyIdx;

                KeyIdx = *(NDIS_802_11_KEY_INDEX *) pData;
                if (KeyIdx & 0x80000000) {
                    NdisStatus = NDIS_STATUS_FAILURE;
                    ERR("OID_802_11_REMOVE_WEP: Invalid KeyIdx 0x%lx!!\n",
                        KeyIdx);
                } else {
                    KeyIdx = KeyIdx & 0x0fffffff;
                    if (KeyIdx >= 4) {
                        NdisStatus = NDIS_STATUS_FAILURE;
                        ERR("OID_802_11_REMOVE_WEP: Invalid KeyIdx 0x%lx!!\n",
                            KeyIdx);
                    } else {
                        pAd->SharedKey[KeyIdx].KeyLen = 0;
                        pAd->SharedKey[KeyIdx].CipherAlg = CIPHER_NONE;
                        AsicRemoveSharedKeyEntry(pAd, 0, (UCHAR) KeyIdx);
                        DBUG(4, "CMDHandler::OID_802_11_REMOVE_WEP (KeyIdx=%d)\n",
                                 KeyIdx);
                    }
                }
            }
            break;

        case OID_802_11_ADD_KEY_WEP:
            {
                PNDIS_802_11_KEY pKey;
                ULONG i, KeyIdx;

                pKey = (PNDIS_802_11_KEY) pData;
                KeyIdx = pKey->KeyIndex & 0x0fffffff;

                // it is a shared key
                if (KeyIdx >= 4) {
                    NdisStatus = NDIS_STATUS_FAILURE;
                    ERR("OID_802_11_ADD_KEY_WEP: Invalid KeyIdx 0x%lx!!\n",
                        KeyIdx);
                } else {
                    UCHAR CipherAlg;

                    pAd->SharedKey[KeyIdx].KeyLen =
                        (UCHAR) pKey->KeyLength;
                    memcpy(pAd->SharedKey[KeyIdx].Key, &pKey->KeyMaterial,
                           pKey->KeyLength);

                    if (pKey->KeyLength == 5)
                        CipherAlg = CIPHER_WEP64;
                    else
                        CipherAlg = CIPHER_WEP128;

                    // always expand the KEY to 16-byte here for efficiency sake. so that in case CKIP is used
                    // sometime later we don't have to do key expansion for each TX in RTUSBHardTransmit().
                    // However, we shouldn't change pAd->SharedKey[BSS0][KeyIdx].KeyLen
                    if (pKey->KeyLength < 16) {
                        for (i = 1; i < (16 / pKey->KeyLength); i++) {
                            memcpy(&pAd->SharedKey[KeyIdx].
                                   Key[i * pKey->KeyLength],
                                   &pKey->KeyMaterial[0], pKey->KeyLength);
                        }
                        memcpy(&pAd->SharedKey[KeyIdx].
                               Key[i * pKey->KeyLength],
                               &pKey->KeyMaterial[0],
                               16 - (i * pKey->KeyLength));
                    }

                    pAd->SharedKey[KeyIdx].CipherAlg = CipherAlg;
                    if (pKey->KeyIndex & 0x80000000) {
                        // Default key for tx (shared key)
                        pAd->PortCfg.DefaultKeyId = (UCHAR) KeyIdx;
                        DBUG(4, "GOTCHA: DefaultKeyId = %d\n", pAd->PortCfg.DefaultKeyId );
                    }

                    DBUG(4, "CMDHandler::OID_802_11_ADD_KEY_WEP (KeyIdx=%d, KeyLen=%d, CipherAlg=%d)\n",
                             pAd->PortCfg.DefaultKeyId,
                             pAd->SharedKey[KeyIdx].KeyLen,
                             pAd->SharedKey[KeyIdx].CipherAlg);
                    AsicAddSharedKeyEntry(pAd, 0, (UCHAR) KeyIdx, CipherAlg,
                                          pAd->SharedKey[KeyIdx].Key, NULL, NULL);
                }
            }
            break;

        case OID_802_11_ADD_KEY:
            {
                PNDIS_802_11_KEY pkey = (PNDIS_802_11_KEY) pData;

                NdisStatus = RTMPWPAAddKeyProc(pAd, pkey);
                RTUSBBulkReceive(pAd);
                DBUG(4, "CMDHandler::OID_802_11_ADD_KEY\n");
            }
            break;

        case OID_802_11_REMOVE_KEY:
            {
                PNDIS_802_11_REMOVE_KEY pRemoveKey;
                ULONG KeyIdx;

                pRemoveKey = (PNDIS_802_11_REMOVE_KEY) pData;
                if (pAd->PortCfg.AuthMode >= Ndis802_11AuthModeWPA) {
                    NdisStatus = RTMPWPARemoveKeyProc(pAd, pData);
                    DBUG(4,"CMDHandler::RTMPWPARemoveKeyProc\n");
                } else {
                    KeyIdx = pRemoveKey->KeyIndex;

                    if (KeyIdx & 0x80000000) {
                        // Should never set default bit when remove key
                        NdisStatus = NDIS_STATUS_FAILURE;
                        ERR("OID_802_11_REMOVE_KEY: Invalid KeyIdx 0x%lx!!\n",
                            KeyIdx);
                    } else {
                        KeyIdx = KeyIdx & 0x0fffffff;
                        if (KeyIdx >= 4) {
                            NdisStatus = NDIS_STATUS_FAILURE;
                            ERR("OID_802_11_REMOVE_KEY: Invalid KeyIdx 0x%lx!!\n",
                                KeyIdx);
                        } else {
                            pAd->SharedKey[KeyIdx].KeyLen = 0;
                            pAd->SharedKey[KeyIdx].CipherAlg = CIPHER_NONE;
                            AsicRemoveSharedKeyEntry(pAd, 0,
                                                     (UCHAR) KeyIdx);
                            DBUG(4, "CMDHandler::AsicRemoveSharedKeyEntry(KeyIdx=%d)\n",
                                     KeyIdx);
                        }
                    }
                }
            }
            break;

        case OID_802_11_POWER_MODE:
            {
                NDIS_802_11_POWER_MODE PowerMode =
                    *(PNDIS_802_11_POWER_MODE) pData;
                DBUG(4, "CMDHandler::OID_802_11_POWER_MODE (=%d)\n",
                         PowerMode);

                // save user's policy here, but not change PortCfg.Psm immediately
                if (PowerMode == Ndis802_11PowerModeCAM) {
                    // clear PSM bit immediately
                    MlmeSetPsmBit(pAd, PWR_ACTIVE);

                    OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
                    if (pAd->PortCfg.bWindowsACCAMEnable == FALSE)
                        pAd->PortCfg.WindowsPowerMode = PowerMode;
                    pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
                } else if (PowerMode == Ndis802_11PowerModeMAX_PSP) {
                    // do NOT turn on PSM bit here, wait until MlmeCheckPsmChange()
                    // to exclude certain situations.
                    //     MlmeSetPsmBit(pAd, PWR_SAVE);
                    if (pAd->PortCfg.bWindowsACCAMEnable == FALSE)
                        pAd->PortCfg.WindowsPowerMode = PowerMode;
                    pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
                    OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
                    pAd->PortCfg.DefaultListenCount = 5;
                } else if (PowerMode == Ndis802_11PowerModeFast_PSP) {
                    // do NOT turn on PSM bit here, wait until MlmeCheckPsmChange()
                    // to exclude certain situations.
                    //     MlmeSetPsmBit(pAd, PWR_SAVE);
                    OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
                    if (pAd->PortCfg.bWindowsACCAMEnable == FALSE)
                        pAd->PortCfg.WindowsPowerMode = PowerMode;
                    pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
                    pAd->PortCfg.DefaultListenCount = 3;
                }
            }
            break;

        case RT_PERFORM_SOFT_DIVERSITY:
            AsicRxAntEvalAction(pAd);
            break;

        case RT_OID_FORCE_WAKE_UP:
            AsicForceWakeup(pAd);
            break;

        case RT_OID_SET_PSM_BIT_ACTIVE:
            MlmeSetPsmBit(pAd, PWR_ACTIVE);
            break;

        default:
            break;
        }

        if (cmdqelmt->CmdFromNdis == TRUE) {
            if ((cmdqelmt->command != OID_802_11_BSSID_LIST_SCAN) &&
                (cmdqelmt->command != RT_OID_802_11_BSSID) &&
                (cmdqelmt->command != OID_802_11_SSID) &&
                (cmdqelmt->command != OID_802_11_DISASSOCIATE)) {
            }

            if ((cmdqelmt->command != RT_OID_MULTI_READ_MAC) &&
                (cmdqelmt->command != RT_OID_VENDOR_READ_BBP) &&
#ifdef RT73_DBG
                (cmdqelmt->command !=
                 RT_OID_802_11_QUERY_HARDWARE_REGISTER) &&
#endif
                (cmdqelmt->command != RT_OID_USB_VENDOR_EEPROM_READ)) {
                if (cmdqelmt->buffer != NULL)
                    kfree(cmdqelmt->buffer);
            }
            if (cmdqelmt != NULL)
                kfree((PCmdQElmt) cmdqelmt);
        } else {
            cmdqelmt->InUse = FALSE;
        }
    }
}

#ifdef RT2X00DEBUGFS
/*
 * Ethtool handlers.
 */
#define CSR_REG_BASE			0x3000
#define CSR_REG_SIZE			0x04b0
#define EEPROM_BASE			0x0000
#define EEPROM_SIZE			0x0100
#define BBP_SIZE			0x0080


static void rt73usb_read_csr(void *dev, const unsigned long word,
                             void *data)
{
    RTMP_ADAPTER *pAd = dev;

    RTUSBReadMACRegister(pAd, CSR_REG_BASE + (word * sizeof(u32)),
                         (u32 *) data);
}

static void rt73usb_write_csr(void *dev, const unsigned long word,
                              void *data)
{
    RTMP_ADAPTER *pAd = dev;

    RTUSBWriteMACRegister(pAd, word, *((u32 *) data));
}

static void rt73usb_read_eeprom(void *dev, const unsigned long word,
                                void *data)
{
    RTMP_ADAPTER *pAd = dev;

    RTUSBReadEEPROM(pAd, word * sizeof(u16), ((u8 *) data), 2);
}

static void rt73usb_write_eeprom(void *dev, const unsigned long word,
                                 void *data)
{
    /* DANGEROUS, DON'T DO THIS! */
}

static void rt73usb_read_bbp(void *dev, const unsigned long word,
                             void *data)
{
    RTMP_ADAPTER *pAd = dev;

    RTUSBReadBBPRegister(pAd, (u8) word, ((u8 *) data));
}

static void rt73usb_write_bbp(void *dev, const unsigned long word,
                              void *data)
{
    RTMP_ADAPTER *pAd = dev;

    RTUSBWriteBBPRegister(pAd, word, *((u8 *) data));
}

static void rt73usb_open_debugfs(RTMP_ADAPTER * pAd)
{
    struct rt2x00debug *debug = &pAd->debug;

    debug->owner = THIS_MODULE;
    debug->mod_name = DRIVER_NAME;
    debug->mod_version = DRIVER_VERSION;
    debug->reg_csr.read = rt73usb_read_csr;
    debug->reg_csr.write = rt73usb_write_csr;
    debug->reg_csr.word_size = sizeof(u32);
    debug->reg_csr.length = CSR_REG_SIZE;
    debug->reg_eeprom.read = rt73usb_read_eeprom;
    debug->reg_eeprom.write = rt73usb_write_eeprom;
    debug->reg_eeprom.word_size = sizeof(u16);
    debug->reg_eeprom.length = EEPROM_SIZE;
    debug->reg_bbp.read = rt73usb_read_bbp;
    debug->reg_bbp.write = rt73usb_write_bbp;
    debug->reg_bbp.word_size = sizeof(u8);
    debug->reg_bbp.length = BBP_SIZE;
    debug->dev = pAd;

    snprintf(debug->intf_name, sizeof(debug->intf_name),
             "%s", pAd->net_dev->name);

    if (rt2x00debug_register(debug))
        printk(KERN_ERR "Failed to register debug handler.\n");
}

static void rt73usb_close_debugfs(RTMP_ADAPTER * pAd)
{
    rt2x00debug_deregister(&pAd->debug);
}
#else                           /* RT2X00DEBUGFS */
static inline void rt73usb_open_debugfs(RTMP_ADAPTER * pAd)
{
}
static inline void rt73usb_close_debugfs(RTMP_ADAPTER * pAd)
{
}
#endif                          /* RT2X00DEBUGFS */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
extern VOID MlmeWork(struct work_struct *work);
#else
extern VOID MlmeWork(void *vpAd);
#endif


static NDIS_STATUS rt73_load_firmware( struct net_device *net_dev )
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) net_dev->priv;
    ULONG MacCsr0 = 0, Index = 0;

    // Wait for hardware stable
    do {
        Status = RTUSBReadMACRegister(pAd, MAC_CSR0, &MacCsr0);

        DBUG(4, "Init: MAC_CSR0=0x%08x, Status=0x%08x, Index=%d\n",
             MacCsr0, Status, Index);

        if (MacCsr0 != 0)
            break;
        RTMPusecDelay(500); // GRH 2/19/2008 -- changed from one 1000usec
        RTMPusecDelay(500); // delay to two 500usec delays, refer to 
                            // comments in RTMPusecDelay() for details
    } while (Index++ < 1000);

    // Load 8051 firmware
    Status = NICLoadFirmware(pAd);
    if (Status != NDIS_STATUS_SUCCESS) {
        return Status;
    }
    // Initialize Asics
    NICInitializeAsic(pAd);
    return Status;
}


static int usb_rtusb_open(struct net_device *net_dev)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) net_dev->priv;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    UCHAR TmpPhy;

    DBUG_HERE( 4 );
    printk("rt73 driver version - %s\n", DRIVER_VERSION);

    init_MUTEX(&(pAd->usbdev_semaphore));

    // init mediastate to disconnected
    OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED);
    OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_MAX_RETRY_ENABLED); /*GRH 5/6/2007*/ 

    pAd->rx_bh.func = RTUSBRxPacket;
    pAd->rx_bk.func = rtusb_bulkrx;

    // Initialize pAd->PortCfg to manufacture default
    PortCfgInit(pAd);

    // Init  RTMP_ADAPTER CmdQElements
    Status = RTMPInitAdapterBlock(pAd);
    if (Status != NDIS_STATUS_SUCCESS) {
        return Status;
    }
    //
    // Init send data structures and related parameters
    //
    Status = NICInitTransmit(pAd);
    if (Status != NDIS_STATUS_SUCCESS) {
        return Status;
    }
    //
    // Init receive data structures and related parameters
    //
    Status = NICInitRecv(pAd);
    if (Status != NDIS_STATUS_SUCCESS) {
        goto out;
    }

    Status = rt73_load_firmware(net_dev);
    if (Status != NDIS_STATUS_SUCCESS) {
        goto out;
    }
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    INIT_WORK(&pAd->mlme_work, MlmeWork);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
    INIT_WORK(&pAd->mlme_work, MlmeWork, (void *)pAd);
#endif

    // Read RaConfig profile parameters
#ifdef  READ_PROFILE_FROM_FILE
    RTMPReadParametersFromFile(pAd);
#endif

    //
    // Read additional info from NIC such as MAC address
    // This function must called after register CSR base address
    //
#ifdef	INIT_FROM_EEPROM
    NICReadEEPROMParameters(pAd);
    NICInitAsicFromEEPROM(pAd);
#endif
    RTUSBWriteHWMACAddress(pAd);

    // external LNA has different R17 base
    if (pAd->NicConfig2.field.ExternalLNA) {
        pAd->BbpTuning.R17LowerBoundA += 0x10;
        pAd->BbpTuning.R17UpperBoundA += 0x10;
        pAd->BbpTuning.R17LowerBoundG += 0x10;
        pAd->BbpTuning.R17UpperBoundG += 0x10;
    }
    // hardware initialization after all parameters are acquired from
    // Registry or E2PROM
    TmpPhy = pAd->PortCfg.PhyMode;
    pAd->PortCfg.PhyMode = 0xff;
    RTMPSetPhyMode(pAd, TmpPhy);

	// max desired rate might be reset as call phymode setup, so set TxRate again
	if (pAd->PortCfg.DefaultMaxDesiredRate == 0)
		RTMPSetDesiredRates(pAd, -1);
	else
		RTMPSetDesiredRates(pAd, (LONG) (RateIdToMbps[pAd->PortCfg.DefaultMaxDesiredRate - 1] * 1000000));

    //
    // initialize MLME
    //
    Status = MlmeInit(pAd);
    if (Status != NDIS_STATUS_SUCCESS) {
        goto out;
    }
    // mlmethread & RTUSBCmd flag restart
    mlme_kill = 0;
    RTUSBCmd_kill = 0;

    CreateThreads(net_dev);

    // GRH 6/5/2007: version 1.0.3.6 not version 1.0.4.0
    // at every open handler, copy mac address.
    if (!memcmp(net_dev->dev_addr, "\x00\x00\x00\x00\x00\x00", 6))
        memcpy(pAd->net_dev->dev_addr, pAd->CurrentAddress,
               pAd->net_dev->addr_len);
    else
        memcpy(pAd->CurrentAddress, pAd->net_dev->dev_addr,
               pAd->net_dev->addr_len);

    // Clear Reset Flag before starting receiving/transmitting
    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS);
	RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_REMOVE_IN_PROGRESS);

    if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) {
        DBUG( 1, "RADIO ON, ENABLE RECEIVER\n" );
        RTUSBBulkReceive(pAd);
        RTUSBWriteMACRegister(pAd, TXRX_CSR0, 0x025eb032);      // enable RX of MAC block, Staion not drop control frame
        // Initialize RF register to default value
        AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
        AsicLockChannel(pAd, pAd->PortCfg.Channel);
    }

    // USB_ID info for UI
    pAd->VendorDesc = 0x148F2573;

    // Start net_dev interface tx /rx
    netif_start_queue(net_dev);

    netif_carrier_on(net_dev);
    netif_wake_queue(net_dev);
    DBUG(1, "Flags=%x, BulkFlags=%x, OpStatusFlags=%x\n",
         pAd->Flags, pAd->BulkFlags, pAd->PortCfg.OpStatusFlags);
    return 0;

  out:
    ERR("status=%d (but returning 0!)\n", Status);
    ReleaseAdapter(pAd, TRUE, FALSE);
    return -ENODEV;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#else
static int usb_rtusb_close(struct net_device *net_dev)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) net_dev->priv;
    int ret;
    int i = 0;

    DECLARE_WAIT_QUEUE_HEAD(unlink_wakeup);
    DECLARE_WAITQUEUE(wait, current);

    DBUG_HERE( 4 );

    netif_carrier_off(pAd->net_dev);
    netif_stop_queue(pAd->net_dev);

    // ensure there are no more active urbs.
    add_wait_queue(&unlink_wakeup, &wait);
    pAd->wait = &unlink_wakeup;

    // maybe wait for deletions to finish.
    while ((i < 25) && atomic_read(&pAd->PendingRx) > 0) {
#if LINUX_VERSION_CODE >KERNEL_VERSION(2,6,9)
        msleep(UNLINK_TIMEOUT_MS);
#endif
        i++;
    }
    DBUG(4, "pAd->PendingRx=%d, i=%d\n", atomic_read(&pAd->PendingRx), i);

    pAd->wait = NULL;
    remove_wait_queue(&unlink_wakeup, &wait);

    if (pAd->MLMEThr_pid >= 0) {
        mlme_kill = 1;
        RTUSBMlmeUp(pAd);
        wmb();                  // need to check
        ret = kill_proc(pAd->MLMEThr_pid, SIGTERM, 1);
        if (ret) {
            printk(KERN_ERR "%s: unable to signal thread\n",
                   pAd->net_dev->name);
            //return ret;
        }
        wait_for_completion(&pAd->notify);
		// reset mlme thread
		pAd->MLMEThr_pid = -1;
    }
    if (pAd->RTUSBCmdThr_pid >= 0) {
        RTUSBCmd_kill = 1;
        RTUSBCMDUp(pAd);
        wmb();                  // need to check
        ret = kill_proc(pAd->RTUSBCmdThr_pid, SIGTERM, 1);
        if (ret) {
            printk(KERN_ERR "%s: unable to signal thread\n",
                   pAd->net_dev->name);
            //return ret;
        }
        wait_for_completion(&pAd->notify);
		// reset cmd thread
		pAd->RTUSBCmdThr_pid = -1;
    }
    RTUSBHalt(pAd, TRUE);

	RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_REMOVE_IN_PROGRESS);

    DBUG(4, "Flags=%x, BulkFlags=%x, OpStatusFlags=%x\n",
         pAd->Flags, pAd->BulkFlags, pAd->PortCfg.OpStatusFlags);
    DBUG(4, "done.\n");

    return 0;
}

INT MlmeThread(IN void *Context)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) Context;

    DBUG_HERE(4);

    daemonize("rt73");
    allow_signal(SIGTERM);
    current->flags |= PF_NOFREEZE;
    /* signal that we've started the thread */
    complete(&(pAd->notify));

	while (!mlme_kill)
	{
		/* lock the device pointers */
		down(&(pAd->mlme_semaphore));

		down(&(pAd->usbdev_semaphore)); /* global lock? */
		MlmeHandler(pAd);		
		up(&(pAd->usbdev_semaphore));   /* global unlock? */
	}

    /* notify the exit routine that we're actually exiting now
     *
     * complete()/wait_for_completion() is similar to up()/down(),
     * except that complete() is safe in the case where the structure
     * is getting deleted in a parallel mode of execution (i.e. just
     * after the down() -- that's necessary for the thread-shutdown
     * case.
     *
     * complete_and_exit() goes even further than this -- it is safe in
     * the case that the thread of the caller is going away (not just
     * the structure) -- this is necessary for the module-remove case.
     * This is important in preemption kernels, which transfer the flow
     * of execution immediately upon a complete().
     */
    complete_and_exit(&pAd->notify, 0);
    DBUG(4, "done.\n");
}

INT RTUSBCmdThread(IN void *Context)
{
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) Context;

    daemonize("rt73");
    allow_signal(SIGTERM);
    current->flags |= PF_NOFREEZE;
    /* signal that we've started the thread */
    complete(&(pAd->notify));

	while (!RTUSBCmd_kill)
	{
		/* lock the device pointers */
		down(&(pAd->RTUSBCmd_semaphore));

		down(&(pAd->usbdev_semaphore)); /* global lock? */
		CMDHandler(pAd);		
		up(&(pAd->usbdev_semaphore));   /* global unlock? */
	}

    /* notify the exit routine that we're actually exiting now
     *
     * complete()/wait_for_completion() is similar to up()/down(),
     * except that complete() is safe in the case where the structure
     * is getting deleted in a parallel mode of execution (i.e. just
     * after the down() -- that's necessary for the thread-shutdown
     * case.
     *
     * complete_and_exit() goes even further than this -- it is safe in
     * the case that the thread of the caller is going away (not just
     * the structure) -- this is necessary for the module-remove case.
     * This is important in preemption kernels, which transfer the flow
     * of execution immediately upon a complete().
     */
    complete_and_exit(&pAd->notify, 0);
}


////////////////////////////////////////////////////////////////////////////////
#define GRH_PROCFILE
#ifdef GRH_PROCFILE


#ifdef RT73_DBUG_PRINTK
typedef struct _dbug_page_ {
    struct _dbug_page_* next;
    int                 leng;
    char                data[ PAGE_SIZE - 1024/*sizeof(*void) - sizeof(int)*/ ];
} DBUG_PAGE;

static spinlock_t   dbug_lock;
static DBUG_PAGE*   dbug_head;
static DBUG_PAGE*   dbug_tail;


static void dbug_init( void )
{
    NdisAllocateSpinLock(&dbug_lock);
    dbug_head = NULL;
    dbug_tail = NULL;
}


static void _free_first_page( void )
{
    DBUG_PAGE* page;

    page = dbug_head;
    if ( !page ) {
        return;
    }
    dbug_head = dbug_head->next;
    if ( dbug_head == NULL )
        dbug_tail = NULL;

    free_page( (unsigned long) page );
}


static void free_first_page( void )
{
    unsigned long   flags;

    NdisAcquireSpinLock(&dbug_lock, flags);
    _free_first_page();
    NdisReleaseSpinLock(&dbug_lock, flags);
}


static int _number_pages( void )
{
    int count = 0;
    DBUG_PAGE* page;

    for ( page = dbug_head; page != NULL; page = page->next) {
        ++count;
    }
    return count;
}


static int number_pages( void )
{
    unsigned long   flags;
    int             pages;

    NdisAcquireSpinLock(&dbug_lock, flags);
    pages = _number_pages();
    NdisReleaseSpinLock(&dbug_lock, flags);
    return pages;
}


static DBUG_PAGE*   append_new_page( void )
{
    DBUG_PAGE*    page = ( DBUG_PAGE* ) get_zeroed_page( GFP_ATOMIC );
    int           npages;

    // limit the size of the trace buffer based on dbug_pages setting
    for (npages = _number_pages(); npages >= dbug_pages; --npages) {
        _free_first_page();
    }

    if ( page == NULL ) {
        if ( dbug_head == NULL )
            return NULL;
        if ( dbug_head == dbug_tail ) {
            memset( dbug_head, 0, PAGE_SIZE );
            return dbug_head;
        }
        page = dbug_head;
        dbug_head = dbug_head->next;
        if ( dbug_head == NULL )
            dbug_tail = NULL;

        memset( page, 0, PAGE_SIZE );
    }

    if ( dbug_head == NULL ) {
        dbug_tail = page;
        dbug_head = page;
    } else {
        dbug_tail->next = page;
        dbug_tail = page;
    }

    return page;
}


static void   append_new_data( char* text, int leng )
{
    unsigned long   flags;

    NdisAcquireSpinLock(&dbug_lock, flags);
    do {
        if ( dbug_tail && dbug_tail->leng + leng <= sizeof(dbug_tail->data) ) {
            memcpy( dbug_tail->data + dbug_tail->leng, text, leng );
            dbug_tail->leng += leng;
            break;
        }
        append_new_page();
    } while ( dbug_tail != NULL );
    NdisReleaseSpinLock(&dbug_lock, flags);
}


int dbug_printk( const char* fmt, ... )
{
    char     text[ 256 ];
    int      leng;
    va_list  args;

    va_start(args, fmt);
    leng = vsnprintf( text, sizeof(text), fmt, args );
    va_end(args);
    append_new_data( text, leng );
    return leng;
}
#endif


/***********************************************************************
 * Proc FileSystem Routines
 ***********************************************************************/

static struct proc_dir_entry* __init proc_file( 
         char* name, 
         struct proc_dir_entry* parent,
         int (*read_proc)( char* buf, char** start, off_t offset, 
                           int count, int* eof, void* data ),
         int (*write_proc)( struct file* file, const char* buf,
                            unsigned long count, void* data ),
         void* data )
{
    struct proc_dir_entry* entry;

    entry = create_proc_entry( name, 0, parent );
    if ( entry ) {
        entry->read_proc = read_proc;
        entry->write_proc = write_proc;
        entry->data = data;
    }

    return entry;
}


static int getval_from_user( struct file* file, 
                             const char* buf, 
                             u32 count, u32* value )
{
    int   len = count;
    char  string[32];

    if ( len > sizeof(string)-1 ) 
        len = sizeof(string)-1;
    if ( copy_from_user( string, buf, len ) )
        return -EFAULT;

    string[len] = '\0';
    *value = simple_strtoul( string, NULL, 0 );
    return 0;
}


static int proc_get_pAd_cb( char* buf, char** start, off_t offset,
                            int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    leng = 0;

    leng += sprintf( buf + leng, "EAPOLVer=%d\n", d->EAPOLVer );
    leng += sprintf( buf + leng, "MLMEThr_pid=%d\n", d->MLMEThr_pid );
    leng += sprintf( buf + leng, "RTUSBCmdThr_pid=%d\n", d->RTUSBCmdThr_pid );
    leng += sprintf( buf + leng, "bLocalAdminMAC=%d\n", d->bLocalAdminMAC );
    leng += sprintf( buf + leng, "bAcceptDirect=%d\n", d->bAcceptDirect );
    leng += sprintf( buf + leng, "bAcceptMulticast=%d\n", d->bAcceptMulticast );
    leng += sprintf( buf + leng, "bAcceptAllMulticast=%d\n", d->bAcceptAllMulticast );
    leng += sprintf( buf + leng, "bAcceptRFMONTx=%d\n", d->bAcceptRFMONTx );
    leng += sprintf( buf + leng, "NumberOfMcastAddresses=%ld\n", d->NumberOfMcastAddresses );
    leng += sprintf( buf + leng, "TupleCacheLastUpdateIndex=%d\n", d->TupleCacheLastUpdateIndex );
    leng += sprintf( buf + leng, "Flags=0x%lx\n", d->Flags );
    leng += sprintf( buf + leng, "Sequence=%d\n", d->Sequence );
    leng += sprintf( buf + leng, "LinkDownTime=%ld\n", d->LinkDownTime );
    leng += sprintf( buf + leng, "LastRxRate=%ld\n", d->LastRxRate );
    leng += sprintf( buf + leng, "bConfigChanged=%d\n", d->bConfigChanged );
    leng += sprintf( buf + leng, "RfIcType=%d\n", d->RfIcType );
    leng += sprintf( buf + leng, "RfFreqOffset=%ld\n", d->RfFreqOffset );
    leng += sprintf( buf + leng, "bAutoTxAgc=%d\n", d->bAutoTxAgc );
    leng += sprintf( buf + leng, "bAutoTxAgcA=%d\n", d->bAutoTxAgcA );
    leng += sprintf( buf + leng, "bAutoTxAgcG=%d\n", d->bAutoTxAgcG );
    leng += sprintf( buf + leng, "BulkOutMaxPacketSize=%d\n", d->BulkOutMaxPacketSize );
    leng += sprintf( buf + leng, "BulkInMaxPacketSize=%d\n", d->BulkInMaxPacketSize );
    leng += sprintf( buf + leng, "DeQueueRunning[0]=%d\n", d->DeQueueRunning[0] );
    leng += sprintf( buf + leng, "DeQueueRunning[1]=%d\n", d->DeQueueRunning[1] );
    leng += sprintf( buf + leng, "DeQueueRunning[2]=%d\n", d->DeQueueRunning[2] );
    leng += sprintf( buf + leng, "DeQueueRunning[3]=%d\n", d->DeQueueRunning[3] );
    leng += sprintf( buf + leng, "TxRingTotalNumber[0]=%d\n", d->TxRingTotalNumber[0] );
    leng += sprintf( buf + leng, "TxRingTotalNumber[1]=%d\n", d->TxRingTotalNumber[1] );
    leng += sprintf( buf + leng, "TxRingTotalNumber[2]=%d\n", d->TxRingTotalNumber[2] );
    leng += sprintf( buf + leng, "TxRingTotalNumber[3]=%d\n", d->TxRingTotalNumber[3] );
    leng += sprintf( buf + leng, "NextTxIndex[0]=%d\n", d->NextTxIndex[0] );
    leng += sprintf( buf + leng, "NextTxIndex[1]=%d\n", d->NextTxIndex[1] );
    leng += sprintf( buf + leng, "NextTxIndex[2]=%d\n", d->NextTxIndex[2] );
    leng += sprintf( buf + leng, "NextTxIndex[3]=%d\n", d->NextTxIndex[3] );
    leng += sprintf( buf + leng, "NextMLMEIndex=%d\n", d->NextMLMEIndex );
    leng += sprintf( buf + leng, "PushMgmtIndex=%d\n", d->PushMgmtIndex );
    leng += sprintf( buf + leng, "PopMgmtIndex=%d\n", d->PopMgmtIndex );
    leng += sprintf( buf + leng, "NextRxBulkInIndex=%d\n", d->NextRxBulkInIndex );
    leng += sprintf( buf + leng, "NextBulkOutIndex[0]=%d\n", d->NextBulkOutIndex[0] );
    leng += sprintf( buf + leng, "NextBulkOutIndex[1]=%d\n", d->NextBulkOutIndex[1] );
    leng += sprintf( buf + leng, "NextBulkOutIndex[2]=%d\n", d->NextBulkOutIndex[2] );
    leng += sprintf( buf + leng, "NextBulkOutIndex[3]=%d\n", d->NextBulkOutIndex[3] );
    leng += sprintf( buf + leng, "BulkOutPending[0]=%d\n", d->BulkOutPending[0] );
    leng += sprintf( buf + leng, "BulkOutPending[1]=%d\n", d->BulkOutPending[1] );
    leng += sprintf( buf + leng, "BulkOutPending[2]=%d\n", d->BulkOutPending[2] );
    leng += sprintf( buf + leng, "BulkOutPending[3]=%d\n", d->BulkOutPending[3] );
    leng += sprintf( buf + leng, "PrioRingTxCnt=%ld\n", d->PrioRingTxCnt );
    leng += sprintf( buf + leng, "PrioRingFirstIndex=%d\n", d->PrioRingFirstIndex );
    leng += sprintf( buf + leng, "TxCount=%d\n", atomic_read( &d->TxCount ) );
    leng += sprintf( buf + leng, "PendingTx=%d\n", d->PendingTx );
    leng += sprintf( buf + leng, "NumPacketsQueued[0]=%d\n", d->NumPacketsQueued[0] );
    leng += sprintf( buf + leng, "NumPacketsQueued[1]=%d\n", d->NumPacketsQueued[1] );
    leng += sprintf( buf + leng, "NumPacketsQueued[2]=%d\n", d->NumPacketsQueued[2] );
    leng += sprintf( buf + leng, "NumPacketsQueued[3]=%d\n", d->NumPacketsQueued[3] );
    leng += sprintf( buf + leng, "PendingRx=%d\n", atomic_read( &d->PendingRx ) );
    leng += sprintf( buf + leng, "BulkFlags=%ld\n", d->BulkFlags );
    leng += sprintf( buf + leng, "BulkOutDataOneSecCount=%ld\n", d->BulkOutDataOneSecCount );
    leng += sprintf( buf + leng, "BulkInDataOneSecCount=%ld\n", d->BulkInDataOneSecCount );
    leng += sprintf( buf + leng, "BulkLastOneSecCount=%ld\n", d->BulkLastOneSecCount );
    return leng;
}


static int proc_get_port_config_cb( char* buf, char** start, off_t offset,
                                    int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    i, leng = 0;

    leng += sprintf( buf + leng, "Psm=%d DisassocReason=%d DisassocSta[%02x:%02x:%02x:%02x:%02x:%02x]\n",
                     d->PortCfg.Psm, d->PortCfg.DisassocReason,
                     d->PortCfg.DisassocSta[0],
                     d->PortCfg.DisassocSta[1],
                     d->PortCfg.DisassocSta[2],
                     d->PortCfg.DisassocSta[3],
                     d->PortCfg.DisassocSta[4],
                     d->PortCfg.DisassocSta[5] );
    leng += sprintf( buf + leng, "DeathReason=%d DeauthSta[%02x:%02x:%02x:%02x:%02x:%02x]\n",
                     d->PortCfg.DeauthReason,
                     d->PortCfg.DeauthSta[0],
                     d->PortCfg.DeauthSta[1],
                     d->PortCfg.DeauthSta[2],
                     d->PortCfg.DeauthSta[3],
                     d->PortCfg.DeauthSta[4],
                     d->PortCfg.DeauthSta[5] );
    leng += sprintf( buf + leng, "AuthFailReason=%d AuthFailSta[%02x:%02x:%02x:%02x:%02x:%02x]\n",
                     d->PortCfg.AuthFailReason,
                     d->PortCfg.AuthFailSta[0],
                     d->PortCfg.AuthFailSta[1],
                     d->PortCfg.AuthFailSta[2],
                     d->PortCfg.AuthFailSta[3],
                     d->PortCfg.AuthFailSta[4],
                     d->PortCfg.AuthFailSta[5] );
    leng += sprintf( buf + leng, "AuthMode=%s WepStatus=%s OrigWepStatus=%s GroupCipher=%s PairCipher=%s\n",
                     GetAuthMode(d->PortCfg.AuthMode, 0),
                     GetEncryptType(d->PortCfg.WepStatus, 0),
                     GetEncryptType(d->PortCfg.OrigWepStatus, 0),
                     GetEncryptType(d->PortCfg.GroupCipher, 0),
                     GetEncryptType(d->PortCfg.PairCipher, 0) );
    leng += sprintf( buf + leng, "bMixCipher=%d RsnCapability=0x%x PortSecured=%d RSN_IELen=%d IEEE8021X=%d\n",
                     d->PortCfg.bMixCipher, d->PortCfg.RsnCapability, d->PortCfg.PortSecured,
                     d->PortCfg.RSN_IELen, d->PortCfg.IEEE8021X );
    leng += sprintf( buf + leng, "IEEE8021X_required_keys=%d WPA_Supplicant=%d\n",
                     d->PortCfg.IEEE8021x_required_keys, d->PortCfg.WPA_Supplicant);
    leng += sprintf( buf + leng, "bBlockAssoc=%d WpaState=%d DefaultKeyId=%d Channel=%d AdhocChannel=%d\n",
                     d->PortCfg.bBlockAssoc, d->PortCfg.WpaState, d->PortCfg.DefaultKeyId, d->PortCfg.Channel, d->PortCfg.AdhocChannel );
    for (i=0; i<d->PortCfg.SupRateLen; ++i) {
        if (i >= MAX_LEN_OF_SUPPORTED_RATES) {
            break;
        }
        leng += sprintf( buf + leng, "SupRate[%d]=%d\n", i, d->PortCfg.SupRate[i] );
    }
    for (i=0; i<d->PortCfg.ExtRateLen; ++i) {
        if (i >= MAX_LEN_OF_SUPPORTED_RATES) {
            break;
        }
        leng += sprintf( buf + leng, "ExtRate[%d]=%d\n", i, d->PortCfg.ExtRate[i] );
    }
    leng += sprintf( buf + leng, "BasicRateBitmap=0x%lx MaxDesireRate=%d BasicMlmeRate=%d\n",
                     d->PortCfg.BasicRateBitmap, d->PortCfg.MaxDesiredRate, d->PortCfg.BasicMlmeRate );
    leng += sprintf( buf + leng, "Bssid[%02x:%02x:%02x:%02x:%02x:%02x] LastBssid[%02x:%02x:%02x:%02x:%02x:%02x]\n",
                     d->PortCfg.Bssid[0],
                     d->PortCfg.Bssid[1],
                     d->PortCfg.Bssid[2],
                     d->PortCfg.Bssid[3],
                     d->PortCfg.Bssid[4],
                     d->PortCfg.Bssid[5],
                     d->PortCfg.LastBssid[0],
                     d->PortCfg.LastBssid[1],
                     d->PortCfg.LastBssid[2],
                     d->PortCfg.LastBssid[3],
                     d->PortCfg.LastBssid[4],
                     d->PortCfg.LastBssid[5] );
    leng += sprintf( buf + leng, "%s", "Ssid=<" );
    for (i=0;i<d->PortCfg.SsidLen;++i) {
        if (i >= MAX_LEN_OF_SSID) {
            break;
        }
        leng += sprintf( buf + leng, "%c", d->PortCfg.Ssid[i] );
    }
    leng += sprintf( buf + leng, "%s", "> LastSsid=<" );

    for (i=0;i<d->PortCfg.LastSsidLen;++i) {
        if (i >= MAX_LEN_OF_SSID) {
            break;
        }
        leng += sprintf( buf + leng, "%c", d->PortCfg.LastSsid[i] );
    }
    leng += sprintf( buf + leng, ">\n" );

    leng += sprintf( buf + leng, "BssType=%d AtimWin=%d bAutoReconnect=%d PhyMode=%d AdHocMode=%ld\n",
                     d->PortCfg.BssType, d->PortCfg.AtimWin, d->PortCfg.bAutoReconnect, 
                     d->PortCfg.PhyMode, d->PortCfg.AdhocMode );
    leng += sprintf( buf + leng, "bFastRoaming=%d dBmToRoam=%ld bGetAPConfig=%d bWscCapable=%d UseShortSlotTime=%ld\n",
                     d->PortCfg.bFastRoaming, d->PortCfg.dBmToRoam, d->PortCfg.bGetAPConfig, d->PortCfg.bWscCapable,
                     d->PortCfg.UseShortSlotTime );
    return leng;
}


static int proc_get_active_config_cb( char* buf, char** start, off_t offset,
                                       int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    i, leng = 0;

    leng += sprintf( buf + leng, "Aid=%d AtimWin=%d CapabilityInfo=0x%x CfpMaxDuration=%d CfpPeriod=%d SupRateLen=%d ExtRateLen=%d\n",
                     d->ActiveCfg.Aid, d->ActiveCfg.AtimWin, d->ActiveCfg.CapabilityInfo, d->ActiveCfg.CfpMaxDuration, d->ActiveCfg.CfpPeriod,
                     d->ActiveCfg.SupRateLen, d->ActiveCfg.ExtRateLen );
    for (i=0; i<d->ActiveCfg.SupRateLen; ++i) {
        if (i >= MAX_LEN_OF_SUPPORTED_RATES) {
            break;
        }
        leng += sprintf( buf + leng, "SupRate[%d]=%d\n", i, d->ActiveCfg.SupRate[i] );
    }
    for (i=0; i<d->ActiveCfg.ExtRateLen; ++i) {
        if (i >= MAX_LEN_OF_SUPPORTED_RATES) {
            break;
        }
        leng += sprintf( buf + leng, "ExtRate[%d]=%d\n", i, d->ActiveCfg.ExtRate[i] );
    }

    return leng;
}


static int proc_get_counters_cb( char* buf, char** start, off_t offset,
                                 int count, int* eof, void* data )
{
    //PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    leng = 0;
    return leng;
}


static int proc_get_dev_stats_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    leng = 0;

    leng += sprintf( buf + leng, "rx_packets=%ld tx_packets=%ld rx_bytes=%ld tx_bytes=%ld\n",
                d->WlanCounters.ReceivedFragmentCount.vv.LowPart, // total packets received
                d->WlanCounters.TransmittedFragmentCount.vv.LowPart,      // total packets transmitted
                d->RalinkCounters.ReceivedByteCount,        // total bytes received
                d->RalinkCounters.TransmittedByteCount );    // total bytes transmitted );
    leng += sprintf( buf + leng, "rx_errors=%ld tx_errors=%ld rx_dropped=%ld tx_dropped=%ld multicast=%ld\n",
                d->Counters8023.RxErrors,  // bad packets received
                d->Counters8023.TxErrors,  // packet transmit problems
                d->Counters8023.RxNoBuffer,       // no space in linux buffers
                d->WlanCounters.FailedCount.vv.LowPart,   // no space available in linux
                d->WlanCounters.MulticastReceivedFrameCount.vv.LowPart );    // multicast packets received
    leng += sprintf( buf + leng, "collisions=%ld rx_over_errors=%ld rx_frame_errors=%ld rx_fifo_errors=%ld\n",
                d->Counters8023.OneCollision + d->Counters8023.MoreCollisions,  // Collision packets
                d->Counters8023.RxNoBuffer,   // receiver ring buff overflow
                d->Counters8023.RcvAlignmentErrors,  // recv'd frame alignment error
                d->Counters8023.RxNoBuffer );   // recv'r fifo overrun

    return leng;
}


static int proc_get_mlme_struct_cb( char* buf, char** start, off_t offset,
                                    int count, int* eof, void* data )
{
    //PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    leng = 0;
    return leng;
}


static int proc_get_mlme_aux_cb( char* buf, char** start, off_t offset,
                                 int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    i, leng = 0;

    leng += sprintf( buf + leng, "BssType=%d Bssid=[%02x:%02x:%02x:%02x:%02x:%02x] Ssid=<",
                     d->MlmeAux.BssType,
                     d->MlmeAux.Bssid[0],
                     d->MlmeAux.Bssid[1],
                     d->MlmeAux.Bssid[2],
                     d->MlmeAux.Bssid[3],
                     d->MlmeAux.Bssid[4],
                     d->MlmeAux.Bssid[5] );
    for (i=0;i<d->MlmeAux.SsidLen;++i) {
        if (i >= MAX_LEN_OF_SSID) {
            break;
        }
        leng += sprintf( buf + leng, "%c", d->MlmeAux.Ssid[i] );
    }
    leng += sprintf( buf + leng, "%s", "> AutoReconnectSsid=<" );

    for (i=0;i<d->MlmeAux.AutoReconnectSsidLen;++i) {
        if (i >= MAX_LEN_OF_SSID) {
            break;
        }
        leng += sprintf( buf + leng, "%c", d->MlmeAux.AutoReconnectSsid[i] );
    }
    leng += sprintf( buf + leng, ">\n" );

    leng += sprintf( buf + leng, "Alg=%d ScanType=%d Channel=%d Aid=%d CapInfo=0x%x\n",
                     d->MlmeAux.Alg,
                     d->MlmeAux.ScanType,
                     d->MlmeAux.Channel,
                     d->MlmeAux.Aid,
                     d->MlmeAux.CapabilityInfo );
    leng += sprintf( buf + leng, "BeaconPeriod=%d BssIdx=%ld RoamIdx=%ld CurrReqIsFromNdis=%d\n",
                     d->MlmeAux.BeaconPeriod,
                     d->MlmeAux.BssIdx,
                     d->MlmeAux.RoamIdx,
                     d->MlmeAux.CurrReqIsFromNdis );

    //USHORT CfpMaxDuration;
    //USHORT CfpPeriod;
    //USHORT AtimWin;

    // Copy supported rate from desired AP's beacon. We are trying to match
    // AP's supported and extended rate settings.
    //UCHAR SupRate[MAX_LEN_OF_SUPPORTED_RATES];
    //UCHAR ExtRate[MAX_LEN_OF_SUPPORTED_RATES];
    //UCHAR SupRateLen;
    //UCHAR ExtRateLen;


    //BSS_TABLE SsidBssTab;       // AP list for the same SSID
    //BSS_TABLE RoamTab;          // AP list eligible for roaming

    return leng;
}


static int proc_get_bss_scantab_cb( char* buf, char** start, off_t offset,
                                    int count, int* eof, void* data )
{
    //PRTMP_ADAPTER d = ( PRTMP_ADAPTER ) data;
    int    leng = 0;
    return leng;
}


static int proc_get_rt73regs_cb( char* buf, char** start, off_t offset,
                                 int count, int* eof, void* pdata )
{
    PRTMP_ADAPTER   d = ( PRTMP_ADAPTER ) pdata;
    int             leng = 0;
    int             reg;
    ULONG           data;

    for ( reg = MAC_CSR0; reg != TXRX_CSR0; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "MAC_CSR%d=%lx\n", (reg - MAC_CSR0)/4, data );
    }
    for ( reg = TXRX_CSR0; reg != PHY_CSR0; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "TXRX_CSR%d=%lx\n", (reg - TXRX_CSR0)/4, data );
    }
    for ( reg = PHY_CSR0; reg != SEC_CSR0; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "PHY_CSR%d=%lx\n", (reg - PHY_CSR0)/4, data );
    }
    for ( reg = SEC_CSR0; reg != STA_CSR0; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "SEC_CSR%d=%lx\n", (reg - SEC_CSR0)/4, data );
    }
    for ( reg = STA_CSR0; reg != QOS_CSR0; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "STA_CSR%d=%lx\n", (reg - STA_CSR0)/4, data );
    }
    for ( reg = QOS_CSR0; reg != QOS_CSR5+4; reg += 4 ) {
        RTUSBReadMACRegister(d, reg, &data);
        leng += sprintf( buf + leng, "QOS_CSR%d=%lx\n", (reg - QOS_CSR0)/4, data );
    }

    RTUSBReadMACRegister(d, AIFSN_CSR, &data);
    leng += sprintf( buf + leng, "AIFSN_CSR=%lx\n", data );
    RTUSBReadMACRegister(d, CWMIN_CSR, &data);
    leng += sprintf( buf + leng, "CWMIN_CSR=%lx\n", data );
    RTUSBReadMACRegister(d, CWMAX_CSR, &data);
    leng += sprintf( buf + leng, "CWMAX_CSR=%lx\n", data );
    RTUSBReadMACRegister(d, AC_TXOP_CSR0, &data);
    leng += sprintf( buf + leng, "AC_TXOP_CSR0=%lx\n", data );
    RTUSBReadMACRegister(d, AC_TXOP_CSR1, &data);
    leng += sprintf( buf + leng, "AC_TXOP_CSR1=%lx\n", data );
    return leng;
}

#ifdef RT73_DBUG_PRINTK
static int proc_get_trace_cb( char* buf, char** start, off_t offset,
                              int count, int* eof, void* data )
{
    DBUG_PAGE*      page;
    unsigned long   flags;
    int             leng;
    off_t           off = offset;

    NdisAcquireSpinLock(&dbug_lock, flags);
    for ( page = dbug_head; page != NULL; page = page->next ) {
        if ( off < page->leng )
            break;
        off -= page->leng;
    }

    if ( page == NULL ) {
        *eof = 1;
        NdisReleaseSpinLock(&dbug_lock, flags);
        return 0;
    }

    leng = page->leng - off;
    if ( leng > count )
        leng = count;
    memcpy( buf, page->data + off, leng );
    NdisReleaseSpinLock(&dbug_lock, flags);
    *start = (char*)leng;
    return leng;
}


static int proc_set_trace_cb( struct file* file, const char* buf, 
                              unsigned long count, void* data )
{
    while ( dbug_head ) {
        free_first_page();
    }
    return count;
}


static int proc_get_debugpages_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    *eof = 1;
    return sprintf( buf, "%d of %d\n", number_pages(), dbug_pages );
}


static int proc_set_debugpages_cb( struct file* file, const char* buf, 
                                   unsigned long count, void* data )
{
    u32  value;
    int  rc = getval_from_user(file,buf,count,&value);
    if ( rc ) 
        return rc;

    dbug_pages = value;
    return count;
}


static int proc_get_debuglevel_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    *eof = 1;
    return sprintf( buf, "%d\n", dbug_level );
}


static int proc_set_debuglevel_cb( struct file* file, const char* buf, 
                                   unsigned long count, void* data )
{
    u32  value;
    int  rc = getval_from_user(file,buf,count,&value);
    if ( rc ) 
        return rc;

    dbug_level = value;
    return count;
}
#endif


static int proc_get_zerocopy_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    *eof = 1;
    return sprintf( buf, "%d\n", zero_copy );
}


static int proc_set_zerocopy_cb( struct file* file, const char* buf, 
                                   unsigned long count, void* data )
{
    u32  value;
    int  rc = getval_from_user(file,buf,count,&value);
    if ( rc ) 
        return rc;

    zero_copy = value;
    return count;
}


static int proc_get_rtdebuglevel_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    *eof = 1;
    return sprintf( buf, "%ld\n", RTDebugLevel );
}


static int proc_set_rtdebuglevel_cb( struct file* file, const char* buf, 
                                   unsigned long count, void* data )
{
    u32  value;
    int  rc = getval_from_user(file,buf,count,&value);
    if ( rc ) 
        return rc;

    RTDebugLevel = value;
    return count;
}


static int proc_get_networkstatus_cb( char* buf, char** start, off_t offset,
                                      int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = (PRTMP_ADAPTER) data;
    int leng = 0;

    *eof = 1;
    if (d == NULL) {
        strcpy(buf + leng, "pAd is NULL\n");
        return strlen(buf);
    }
    leng += sprintf(buf + leng, "MEDIA_CONNECTED=%d\n", 
                    OPSTATUS_TEST_FLAG(d, fOP_STATUS_MEDIA_STATE_CONNECTED));
    leng += sprintf(buf + leng, "ADHOC_ON()=%d\n", ADHOC_ON(d));
    leng += sprintf(buf + leng, "INFRA_ON()=%d\n", INFRA_ON(d));
    leng += sprintf(buf + leng, "ADAPTER_RESET_IN_PROGRESS=%d\n", 
                    RTMP_TEST_FLAG(d, fRTMP_ADAPTER_RESET_IN_PROGRESS));
    leng += sprintf(buf + leng, "ADAPTER_HALT_IN_PROGRESS=%d\n", 
                    RTMP_TEST_FLAG(d, fRTMP_ADAPTER_HALT_IN_PROGRESS));
    return leng;
}




static int proc_get_linkstatus_cb( char* buf, char** start, off_t offset,
                                   int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = (PRTMP_ADAPTER) data;
    int           cnt;
    int           leng = 0;
    int           sex;
    STATUS_ENTRY* se;

    *eof = 1;
    if (d == NULL) {
        return sprintf(buf, "0: STATUS=-2 (link down)\n");
    }

    if (RTMP_TEST_FLAG(d, fRTMP_ADAPTER_RESET_IN_PROGRESS)) {
        return sprintf(buf, "0: STATUS=-3 (RESET in progress)\n");
    }
    if (RTMP_TEST_FLAG(d, fRTMP_ADAPTER_HALT_IN_PROGRESS)) {
        return sprintf(buf, "0: STATUS=-4 (HALT in progress)\n");
    }

    //if (!OPSTATUS_TEST_FLAG(d, fOP_STATUS_MEDIA_STATE_CONNECTED)) {
    //    return sprintf(buf, "0: STATUS=-5 (MEDIA disconnected)\n");
    //}

    cnt = ringcnt(&d->status_ring);
    if (cnt == 0) {
        return sprintf(buf, "0: STATUS=-6 (no link-status entries)\n");
    }

    sex = ringbeg(&d->status_ring);
    while ( leng < count - 48 && cnt != 0 ) {
        int i, keylen, nomedia;

        // if this is the first entry to be printed and the media is not 
        // connected (and the status of the first entry is 0 or -1, change
        // the status to -5 (MEDIA disconnected)
        nomedia = leng == 0 && 
                  !OPSTATUS_TEST_FLAG(d, fOP_STATUS_MEDIA_STATE_CONNECTED);

        se = ringread(&d->status_ring, &sex);
        leng += sprintf( buf + leng, "%d: STATUS=%d NETWORK=%s SSID=<", 
                         cnt--, nomedia && se->status == 0? -5 : se->status,
                         se->BssType == BSS_ADHOC? "ADHOC" : "INFRA" );
        for ( i = 0; i < se->SsidLen; ++i ) {
            if ( i >= sizeof(se->Ssid) )
                break;
            leng += sprintf( buf + leng, "%c", se->Ssid[i] );
        }
        leng += sprintf( buf + leng, "> AUTH=%s ENCRYPT=%s KEY=%s[",
                         GetAuthMode(se->AuthMode, 0), 
                         GetEncryptType(se->WepStatus, 0),
                         CipherName[se->cipher_key.CipherAlg&7] );
        keylen = se->cipher_key.KeyLen;
        if (keylen == 0 && se->AuthMode >= Ndis802_11AuthModeWPA) {
            keylen = 16;
        }
        for ( i = 0; i < keylen; i++ ) {
            if ( i > 0 ) {
                buf[leng++] = ':';
            }
            leng += sprintf( buf + leng, "%02x", se->cipher_key.Key[i] );
        }
        leng += sprintf( buf + leng, "]\n" );
    }
    return leng;
}


static int proc_set_linkstatus_cb( struct file* file, const char* buf, 
                                   unsigned long count, void* data )
{
    PRTMP_ADAPTER d = (PRTMP_ADAPTER) data;
    d->status_ring.next = 0;
    d->status_ring.wrap = 0;
    return count;
}


static int proc_get_txq_full_cb( char* buf, char** start, off_t offset,
                                 int count, int* eof, void* data )
{
    PRTMP_ADAPTER d = (PRTMP_ADAPTER) data;

    *eof = 1;
    if (d == NULL)
        return sprintf(buf, "LINK DOWN!\n");

    return sprintf( buf, "RalinkCounters.MgmtRingFullCount = %ld\n",
                    d->RalinkCounters.MgmtRingFullCount );
}


static int proc_set_txq_full_cb( struct file* file, const char* buf, 
                                 unsigned long count, void* data )
{
    PRTMP_ADAPTER d = (PRTMP_ADAPTER) data;
    printk( "%s(): RalinkCounters.MgmtRingFullCount(%ld) --> 2\n",
            __FUNCTION__, d->RalinkCounters.MgmtRingFullCount );
    d->RalinkCounters.MgmtRingFullCount = 2;
    return count;
}


#define VOID_DIR       -1
#define ROOT_DIR        0
#define WIFI_DIR        1

static struct {
    char* name;
    int   parent;
} procdir[] = {
/*0*/    { "proc-root",     VOID_DIR },
/*1*/    { "chumby-wifi",   ROOT_DIR },
};


static struct {
    char* name;
    int   parent;
    int (*read_proc)( char* buf, char** start, off_t offset, 
                      int count, int* eof, void* data );
    int (*write_proc)( struct file* file, const char* buf,
                       unsigned long count, void* data );
} procfile[] = {
 { "RTDebugLevel",    WIFI_DIR,        proc_get_rtdebuglevel_cb, proc_set_rtdebuglevel_cb },
#ifdef RT73_DBUG_PRINTK
 { "dbug-level",      WIFI_DIR,        proc_get_debuglevel_cb,   proc_set_debuglevel_cb },
 { "dbug-pages",      WIFI_DIR,        proc_get_debugpages_cb,   proc_set_debugpages_cb },
 { "trace-buffer",    WIFI_DIR,        proc_get_trace_cb,        proc_set_trace_cb },
#endif
 { "rtmp-adapter",    WIFI_DIR,        proc_get_pAd_cb,             NULL },
 { "port-config",     WIFI_DIR,        proc_get_port_config_cb,     NULL },
 { "active-config",   WIFI_DIR,        proc_get_active_config_cb,   NULL },
 { "counters",        WIFI_DIR,        proc_get_counters_cb,        NULL },
 { "dev-stats",       WIFI_DIR,        proc_get_dev_stats_cb,       NULL },
 { "mlme-struct",     WIFI_DIR,        proc_get_mlme_struct_cb,     NULL },
 { "mlme-aux",        WIFI_DIR,        proc_get_mlme_aux_cb,        NULL },
 { "bss-scantab",     WIFI_DIR,        proc_get_bss_scantab_cb,     NULL },
 { "regs",            WIFI_DIR,        proc_get_rt73regs_cb,        NULL },
 { "link-status",     WIFI_DIR,        proc_get_linkstatus_cb,      proc_set_linkstatus_cb },
 { "network-status",  WIFI_DIR,        proc_get_networkstatus_cb,   NULL },
 { "zero-copy",       WIFI_DIR,        proc_get_zerocopy_cb,     proc_set_zerocopy_cb },
 { "simulate-txq-full", WIFI_DIR,      proc_get_txq_full_cb,     proc_set_txq_full_cb }
};


static void rt73_proc_rmdir(PRTMP_ADAPTER d)
{
    int i;

    for ( i = ARRAY_SIZE(procdir)-1; i != 0; --i ) {
        if (d->proc_dirs[i]) {
            DBUG(1, "remove_proc_entry(%s)\n", procdir[i].name);
            remove_proc_entry( procdir[i].name, d->proc_dirs[procdir[i].parent] );
            d->proc_dirs[i] = NULL;
        }
    }
}


static void rt73_proc_exit(PRTMP_ADAPTER d)
{
	int i;

    for ( i = 0; i < ARRAY_SIZE(procfile); ++i ) {
        if (d->proc_files[i]) {
            DBUG(1,"remove_proc_entry(%s)\n", procfile[i].name);
            remove_proc_entry( procfile[i].name, d->proc_dirs[procfile[i].parent] );
            d->proc_files[i] = NULL;
        }
    }

    rt73_proc_rmdir(d);
}


static int rt73_proc_mkdir(PRTMP_ADAPTER d)
{
    int i;

    for ( i = 1; i < ARRAY_SIZE(procdir); ++i ) {
        d->proc_dirs[i] = proc_mkdir(procdir[i].name, 
                                     d->proc_dirs[procdir[i].parent]);
        if (d->proc_dirs[i] == NULL) {
            rt73_proc_rmdir(d);
            return -1;
        }
    }

    return 0;
}


static int rt73_proc_init(PRTMP_ADAPTER d)
{
    int i;

    if (ARRAY_SIZE(d->proc_dirs) < ARRAY_SIZE(procdir)) {
        printk( KERN_ERR "%s(): ARRAY_SIZE(d->proc_dirs)[%d] < ARRAY_SIZE(procdir)[%d]\n",
                __FUNCTION__, ARRAY_SIZE(d->proc_dirs), ARRAY_SIZE(procdir) );
        return -1;
    }
    if (ARRAY_SIZE(d->proc_files) < ARRAY_SIZE(procfile)) {
        printk( KERN_ERR "%s(): ARRAY_SIZE(d->proc_files)[%d] < ARRAY_SIZE(procfile)[%d]\n",
                __FUNCTION__, ARRAY_SIZE(d->proc_files), ARRAY_SIZE(procfile) );
        return -1;
    }

    if ( rt73_proc_mkdir(d) )
        return -1;

    for ( i = 0; i < ARRAY_SIZE(procfile); ++i ) {
        struct proc_dir_entry* entry;

        entry = proc_file( procfile[i].name, d->proc_dirs[procfile[i].parent],
                           procfile[i].read_proc, procfile[i].write_proc, d );
        if ( entry == NULL ) {
            rt73_proc_exit(d);
            return -1;
        }

        d->proc_files[i] = entry;
    }

    return 0;
}
#endif
////////////////////////////////////////////////////////////////////////////////


static int usb_rtusb_probe(struct usb_interface *intf,
                           const struct usb_device_id *id)
{
    static RTMP_ADAPTER adapter_memory;
    struct usb_device *dev = interface_to_usbdev(intf);
    PRTMP_ADAPTER pAd;
    int i;
    struct net_device *netdev;
    int res = -ENOMEM;

#ifdef RT73_DBUG_PRINTK
    dbug_init();
#endif
    DBUG_HERE( 4 );
    usb_get_dev(dev);
    for (i = 0; i < rtusb_usb_id_len; i++) {
        if (le32_to_cpu(dev->descriptor.idVendor) ==
            rtusb_usb_id[i].idVendor &&
            le32_to_cpu(dev->descriptor.idProduct) ==
            rtusb_usb_id[i].idProduct) {
            printk("idVendor = 0x%x, idProduct = 0x%x \n",
                   le32_to_cpu(dev->descriptor.idVendor),
                   le32_to_cpu(dev->descriptor.idProduct));
            break;
        }
    }
    if (i == rtusb_usb_id_len) {
        printk("Device Descriptor not matching\n");
        goto out_noalloc;
    }

    /* RTMP_ADAPTER is too big with 64-bit pointers, and using the
     * builtin net_device private area causes the allocation to
     * exceed 128KB and fail.  So we allocate it separately. */
    #ifdef ADAPTER_MEMORY_ON_HEAP
    pAd = kzalloc(sizeof(*pAd), GFP_KERNEL);
    #else
    pAd = &adapter_memory;
    memset( pAd, 0, sizeof(adapter_memory) );
    #endif
    if (!pAd) {
        printk("couldn't allocate RTMP_ADAPTER\n");
        goto out_noalloc;
    }
    DBUG(1,"pAd=%p\n", pAd);
    netdev = alloc_etherdev(0);
    if (!netdev) {
        printk("alloc_etherdev failed\n");
        goto out_nonetdev;
    }

    netdev->priv = pAd;
    pAd->net_dev = netdev;
    netif_stop_queue(netdev);
    pAd->config = &dev->config->desc;
    pAd->pUsb_Dev = dev;
    SET_MODULE_OWNER(netdev);
    ether_setup(netdev);

    netdev->open = usb_rtusb_open;
    netdev->stop = usb_rtusb_close;
    netdev->priv = pAd;
    netdev->hard_start_xmit = RTMPSendPackets;
    netdev->get_stats = rt73_get_ether_stats;

#if WIRELESS_EXT >= 12
#if WIRELESS_EXT < 17
    netdev->get_wireless_stats = rt73_get_wireless_stats;
#endif
    netdev->wireless_handlers =
        (struct iw_handler_def *)&rt73_iw_handler_def;
#endif
    netdev->do_ioctl = rt73_ioctl;

    netdev->hard_header_len = 14;
    netdev->mtu = 1500;
    netdev->addr_len = 6;
    netdev->weight = 64;

    OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED);

    pAd->MLMEThr_pid = -1;
    pAd->RTUSBCmdThr_pid = -1;

    SET_NETDEV_DEV(netdev, &intf->dev);

    {                           // find available
        int i = 0;
        char slot_name[IFNAMSIZ];
        //struct  net_device    *device;
        struct usb_host_interface *iface_desc;
        struct usb_endpoint_descriptor *endpoint;


        for (i = 0; i < 8; i++) {
            sprintf(slot_name, "rausb%d", i);
            if (dev_get_by_name(slot_name) == NULL)
                break;
        }
        if (i == 8) {
            ERR("No available slot name\n");
            return res;
        }

        sprintf(pAd->net_dev->name, "rausb%d", i);
        DBGPRINT(RT_DEBUG_TRACE, "usb device name %s\n",
                 pAd->net_dev->name);


        /* get the active interface descriptor */
        iface_desc = intf->cur_altsetting;

        /* check out the endpoint: it has to be Interrupt & IN */
        endpoint = &iface_desc->endpoint[i].desc;

        /* get Max Packet Size from endpoint */
        pAd->BulkOutMaxPacketSize = (USHORT) endpoint->wMaxPacketSize;
        DBUG(4, "BulkOutMaxPacketSize  %d\n",
                 pAd->BulkOutMaxPacketSize);

    }

    //bottom half data is assign at  each task_scheduler
    //pAd->rx_bh.data = (unsigned long)pAd;
    pAd->rx_bh.func = RTUSBRxPacket;


    res = register_netdev(netdev);
    if (res) {
        printk("register_netdev failed err=%d\n", res);
        goto out;
    }

    usb_set_intfdata(intf, pAd);

    rt73usb_open_debugfs(pAd);

    rt73_proc_init(pAd);
    return 0;

  out:
    free_netdev(netdev);
  out_nonetdev:
    #ifdef ADAPTER_MEMORY_ON_HEAP
    kfree(pAd);
    #endif
  out_noalloc:
    module_put(THIS_MODULE);
    return res;
}

static void usb_rtusb_disconnect(struct usb_interface *intf)
{
    struct usb_device *dev = interface_to_usbdev(intf);
    PRTMP_ADAPTER pAd = (PRTMP_ADAPTER) NULL;

    DBUG_HERE( 4 );
    pAd = usb_get_intfdata(intf);
    usb_set_intfdata(intf, NULL);
    if (!pAd)
    {
        ERR("%s(): pAd is NULL!!!\n", __FUNCTION__ );
        return;
    }

    RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST);
    DBUG(1,"unregister usbnet usb-%s-%s\n",
            dev->bus->bus_name, dev->devpath);

    rt73usb_close_debugfs(pAd);

    tasklet_kill(&pAd->rx_bh);
    tasklet_kill(&pAd->rx_bk);
    // for debug, wait to show some messages to /proc system
    udelay(1);
    //After Add Thread implementation, Upon exec there, pAd->net_dev seems becomes NULL,
    //need to check why???
    //assert(pAd->net_dev != NULL)
    if (pAd->net_dev != NULL) {
        unregister_netdev(pAd->net_dev);
    }
    udelay(1);
    flush_scheduled_work();
    udelay(1);

    rt73_proc_exit(pAd);
    free_netdev(pAd->net_dev);
    #ifdef ADAPTER_MEMORY_ON_HEAP
    kfree(pAd);
    #endif
    usb_put_dev(dev);
    udelay(1);
    DBUG(4,"done.\n");

}
#endif



//
// Driver module load function
//
static INT __init usb_rtusb_init(void)
{
    DBUG_HERE( 4 );
    RTDebugLevel = debug;
    return usb_register(&rtusb_driver);
}

//
// Driver module unload function
//
static VOID __exit usb_rtusb_exit(void)
{
    DBUG_HERE( 4 );
    udelay(1);
    udelay(1);
    usb_deregister(&rtusb_driver);
}

/**************************************/
module_init(usb_rtusb_init);
module_exit(usb_rtusb_exit);
/**************************************/
