/***************************************************************************
 * RT2x00 SourceForge Project - http://rt2x00.serialmonkey.com             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   Licensed under the GNU GPL                                            *
 *   Original code supplied under license from RaLink Inc, 2004.           *
 ***************************************************************************/

/***************************************************************************
 *	Module Name:	rtusb_bulk.c
 *
 *	Abstract:
 *
 *	Revision History:
 *	Who		When		What
 *	--------	----------	-----------------------------
 *	idamlaj		14-10-2006	RFMONTx (based on MarkW's code)
 *
 ***************************************************************************/

#include	"rt_config.h"


static void RTusb_fill_bulk_urb(struct urb *pUrb,
                                struct usb_device *pUsb_Dev,
                                unsigned int bulkpipe,
                                void *pTransferBuf,
                                int BufSize,
                                usb_complete_t Complete, void *pContext)
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
    usb_fill_bulk_urb(pUrb, pUsb_Dev, bulkpipe, pTransferBuf, BufSize,
                      Complete, pContext);
#else
    FILL_BULK_URB(pUrb, pUsb_Dev, bulkpipe, pTransferBuf, BufSize,
                  Complete, pContext);
#endif

}

// ************************ Completion Func ************************ //
VOID RTUSBBulkOutDataPacketComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PTX_CONTEXT pTxContext;
    PRTMP_ADAPTER pAd;
    NTSTATUS status;
    UCHAR BulkOutPipeId;
    unsigned long IrqFlags;

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutDataPacketComplete\n");
    pTxContext = (PTX_CONTEXT) pUrb->context;
    pAd = pTxContext->pAd;
    status = pUrb->status;

    // Store BulkOut PipeId
    BulkOutPipeId = pTxContext->BulkOutPipeId;
    pAd->BulkOutDataOneSecCount++;

    if (status == USB_ST_NOERROR) {
        DBGPRINT_RAW(RT_DEBUG_INFO,
                     "BulkOutDataPacketComplete %d (STATUS_SUCCESS)\n",
                     BulkOutPipeId);

        if (pTxContext->LastOne == TRUE) {
            pAd->Counters.GoodTransmits++;
            FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
            pAd->TxRingTotalNumber[BulkOutPipeId]--;    // sync. to TxCount

            if (!skb_queue_empty(&pAd->SendTxWaitQueue[BulkOutPipeId]))
                RTMPDeQueuePacket(pAd, BulkOutPipeId);
        } else {
            if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS |
                                     fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                     fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                     fRTMP_ADAPTER_BULKOUT_RESET)) {
                FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
                pAd->TxRingTotalNumber[BulkOutPipeId]--;        // sync. to TxCount
                // Indicate next one is frag data which has highest priority
                RTUSB_SET_BULK_FLAG(pAd,
                                    fRTUSB_BULK_OUT_DATA_FRAG << BulkOutPipeId);
            } else {
                while (pTxContext->LastOne != TRUE) {
                    FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
                    pAd->TxRingTotalNumber[BulkOutPipeId]--;    // sync. to TxCount
                    pTxContext = &(pAd->TxContext[BulkOutPipeId]
                                   [pAd->NextBulkOutIndex[BulkOutPipeId]]);
                }

                FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
                pAd->TxRingTotalNumber[BulkOutPipeId]--;        // sync. to TxCount
            }
        }
    }
#if 1                           // STATUS_OTHER
    else {
        ERR("BulkOutPipeId=%d, status=%d\n", BulkOutPipeId, status);

        while (pTxContext->LastOne != TRUE) {
            FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
            pAd->TxRingTotalNumber[BulkOutPipeId]--;    // sync. to TxCount
            pTxContext = &(pAd->TxContext[BulkOutPipeId]
                           [pAd->NextBulkOutIndex[BulkOutPipeId]]);
        }
        FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
        pAd->TxRingTotalNumber[BulkOutPipeId]--;        // sync. to TxCount

        if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS |
                                 fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                 fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                 fRTMP_ADAPTER_BULKOUT_RESET)) {
            RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
            RTUSBEnqueueInternalCmd(pAd, RT_OID_USB_RESET_BULK_OUT);
        }
    }
#endif


    pTxContext =
        &(pAd->TxContext[BulkOutPipeId][pAd->NextBulkOutIndex[BulkOutPipeId]]);
    //
    // bInUse = TRUE, means some process are filling TX data, after that must turn on bWaitingBulkOut
    // bWaitingBulkOut = TRUE, means the TX data are waiting for bulk out.
    //
    if ((pTxContext->bWaitingBulkOut == TRUE) &&
        !RTUSB_TEST_BULK_FLAG(pAd,fRTUSB_BULK_OUT_DATA_FRAG << BulkOutPipeId)) {
        // Indicate There is data avaliable
        RTUSB_SET_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NORMAL << BulkOutPipeId);
    }


    NdisAcquireSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
    pAd->BulkOutPending[BulkOutPipeId] = FALSE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
    // Always call Bulk routine, even reset bulk.
    // The protectioon of rest bulk should be in BulkOut routine
    RTUSBKickBulkOut(pAd);
}

// NULL frame use BulkOutPipeId = 0
VOID RTUSBBulkOutNullFrameComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PRTMP_ADAPTER pAd;
    PTX_CONTEXT pNullContext;
    NTSTATUS status;
    unsigned long IrqFlags;

    pNullContext = (PTX_CONTEXT) pUrb->context;
    pAd = pNullContext->pAd;

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutNullFrameComplete\n");

    // Reset Null frame context flags
    pNullContext->IRPPending = FALSE;
    pNullContext->InUse = FALSE;

    status = pUrb->status;

    if (status == USB_ST_NOERROR) {
        // Don't worry about the queue is empty or not, this function will check itself
        RTMPDeQueuePacket(pAd, 0);
    }
#if 1                           // STATUS_OTHER
    else {
        if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS |
                                 fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                 fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                 fRTMP_ADAPTER_BULKOUT_RESET)) {
            ERR("Bulk Out Null Frame Failed\n");
            RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
            RTUSBEnqueueInternalCmd(pAd, RT_OID_USB_RESET_BULK_OUT);
        }
    }
#endif

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    pAd->BulkOutPending[0] = FALSE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    // Always call Bulk routine, even reset bulk.
    // The protectioon of rest bulk should be in BulkOut routine
    RTUSBKickBulkOut(pAd);

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutNullFrameComplete\n");
}

// RTS frame use BulkOutPipeId = PipeID
VOID RTUSBBulkOutRTSFrameComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PRTMP_ADAPTER pAd;
    PTX_CONTEXT pRTSContext;
    NTSTATUS status;
    unsigned long IrqFlags;

    pRTSContext = (PTX_CONTEXT) pUrb->context;
    pAd = pRTSContext->pAd;

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutRTSFrameComplete\n");

    // Reset RTS frame context flags
    pRTSContext->IRPPending = FALSE;
    pRTSContext->InUse = FALSE;

    status = pUrb->status;

    if (status == USB_ST_NOERROR) {
        // Don't worry about the queue is empty or not, this function will check itself
        RTMPDeQueuePacket(pAd, pRTSContext->BulkOutPipeId);
    }
#if 1                           // STATUS_OTHER
    else {
        if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS | 
                                 fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                 fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                 fRTMP_ADAPTER_BULKOUT_RESET)) {
            ERR("Bulk Out RTS Frame Failed\n");
            RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
            RTUSBEnqueueInternalCmd(pAd, RT_OID_USB_RESET_BULK_OUT);
        }
    }
#endif

    NdisAcquireSpinLock(&pAd->BulkOutLock[pRTSContext->BulkOutPipeId],IrqFlags);
    pAd->BulkOutPending[pRTSContext->BulkOutPipeId] = FALSE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[pRTSContext->BulkOutPipeId],IrqFlags);

    // Always call Bulk routine, even reset bulk.
    // The protectioon of rest bulk should be in BulkOut routine
    RTUSBKickBulkOut(pAd);

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutRTSFrameComplete\n");
}

// MLME use BulkOutPipeId = 0
VOID RTUSBBulkOutMLMEPacketComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PTX_CONTEXT pMLMEContext;
    PRTMP_ADAPTER pAd;
    NTSTATUS status;
    unsigned long IrqFlags;

    pMLMEContext = (PTX_CONTEXT) pUrb->context;
    pAd = pMLMEContext->pAd;
    status = pUrb->status;

    pAd->PrioRingTxCnt--;
    if (pAd->PrioRingTxCnt < 0)
        pAd->PrioRingTxCnt = 0;

    pAd->PrioRingFirstIndex++;
    if (pAd->PrioRingFirstIndex >= PRIO_RING_SIZE) {
        pAd->PrioRingFirstIndex = 0;
    }

    DBGPRINT(RT_DEBUG_INFO,
             "RTUSBBulkOutMLMEPacketComplete::PrioRingFirstIndex = %d, PrioRingTxCnt = %d, PopMgmtIndex = %d, PushMgmtIndex = %d, NextMLMEIndex = %d\n",
             pAd->PrioRingFirstIndex, pAd->PrioRingTxCnt,
             pAd->PopMgmtIndex, pAd->PushMgmtIndex, pAd->NextMLMEIndex);

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutMLMEPacketComplete\n");

    // Reset MLME context flags
    pMLMEContext->IRPPending = FALSE;
    pMLMEContext->InUse = FALSE;
	pMLMEContext->bWaitingBulkOut =FALSE;

    if (status == USB_ST_NOERROR) {
        // Don't worry about the queue is empty or not, this function will check itself
        RTUSBDequeueMLMEPacket(pAd);
    }
#if 1                           // STATUS_OTHER
    else {
        if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS |
                                 fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                 fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                 fRTMP_ADAPTER_BULKOUT_RESET)) {
            ERR("Bulk Out MLME Failed\n");
            RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
            RTUSBEnqueueInternalCmd(pAd, RT_OID_USB_RESET_BULK_OUT);
        }
    }
#endif
	pMLMEContext = &pAd->MLMEContext[pAd->PrioRingFirstIndex];

	if (pAd->PrioRingTxCnt >= 1 && pMLMEContext->bWaitingBulkOut == TRUE)
		RTUSB_SET_BULK_FLAG(pAd, fRTUSB_BULK_OUT_MLME);

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    pAd->BulkOutPending[0] = FALSE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    // Always call Bulk routine, even reset bulk.
    // The protectioon of rest bulk should be in BulkOut routine
    RTUSBKickBulkOut(pAd);

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutMLMEPacketComplete\n");

}

// PS-Poll frame use BulkOutPipeId = 0
VOID RTUSBBulkOutPsPollComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PRTMP_ADAPTER pAd;
    PTX_CONTEXT pPsPollContext;
    NTSTATUS status;
    unsigned long IrqFlags;

    pPsPollContext = (PTX_CONTEXT) pUrb->context;
    pAd = pPsPollContext->pAd;

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutPsPollComplete\n");

    // Reset PsPoll context flags
    pPsPollContext->IRPPending = FALSE;
    pPsPollContext->InUse = FALSE;

    status = pUrb->status;
    if (status == USB_ST_NOERROR) {
        // Don't worry about the queue is empty or not, this function will check itself
        RTMPDeQueuePacket(pAd, 0);
    }
#if 1                           // STATUS_OTHER
    else {
        if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS |
                                 fRTMP_ADAPTER_HALT_IN_PROGRESS  |
                                 fRTMP_ADAPTER_NIC_NOT_EXIST     |
                                 fRTMP_ADAPTER_BULKOUT_RESET)) {
            ERR("Bulk Out PSPoll Failed\n");
            RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
            RTUSBEnqueueInternalCmd(pAd, RT_OID_USB_RESET_BULK_OUT);
        }
    }
#endif

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    pAd->BulkOutPending[0] = FALSE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    // Always call Bulk routine, even reset bulk.
    // The protectioon of rest bulk should be in BulkOut routine
    RTUSBKickBulkOut(pAd);

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutPsPollComplete\n");
}

/*
	========================================================================

	Routine Description:
		This routine process Rx Irp and call rx complete function.

	Arguments:
		DeviceObject	Pointer to the device object for next lower
						device. DeviceObject passed in here belongs to
						the next lower driver in the stack because we
						were invoked via IoCallDriver in USB_RxPacket
						AND it is not OUR device object
	  Irp				Ptr to completed IRP
	  Context			Ptr to our Adapter object (context specified
						in IoSetCompletionRoutine

	Return Value:
		Always returns STATUS_MORE_PROCESSING_REQUIRED

	Note:
		Always returns STATUS_MORE_PROCESSING_REQUIRED
	========================================================================
*/
VOID RTUSBBulkRxComplete(purbb_t pUrb, struct pt_regs *pt_regs)
{
    PRX_CONTEXT pRxContext;
    PRTMP_ADAPTER pAd;
    NTSTATUS status;

    pRxContext = (PRX_CONTEXT) pUrb->context;
    pAd = pRxContext->pAd;

    //
    // We have a number of cases:
    //              1) The USB read timed out and we received no data.
    //              2) The USB read timed out and we received some data.
    //              3) The USB read was successful and fully filled our irp buffer.
    //              4) The irp was cancelled.
    //              5) Some other failure from the USB device object.
    //

    //
    // Free the IRP  and its mdl because they were  alloced by us
    //
    status = pUrb->status;
    atomic_set(&pRxContext->IrpLock, IRPLOCK_COMPLETED);
    atomic_dec(&pAd->PendingRx);

    switch (status) {
    case 0:
        if ((status == USB_ST_NOERROR) &&
            (atomic_read(&pRxContext->IrpLock) != IRPLOCK_CANCE_START)) {
            DBUG(4,"scheduling RTUSBRxPacket(), urb=%p...\n", pUrb);
            //RTUSBRxPacket(pUrb);
            pAd->rx_bh.data = (unsigned long)pUrb;
            tasklet_schedule(&pAd->rx_bh);
            break;

        }                       // STATUS_SUCCESS
        DBUG(11,"status == 0, IrpLock = %d\n",
                 atomic_read(&pRxContext->IrpLock));
        break;

    case -ECONNRESET:          // async unlink
    case -ESHUTDOWN:           // hardware gone = -108
    case -EPROTO:			   // protocol error
    default:                   /*GRH 5/7/2007*/ 
        pUrb = NULL;
        WARN("status = %d\n", status);
        break;
    }
}

static VOID RTUSBInitTxDesc(IN PRTMP_ADAPTER pAd,
                            IN PTX_CONTEXT pTxContext,
                            IN UCHAR BulkOutPipeId, IN usb_complete_t Func)
{
    PURB pUrb = pTxContext->pUrb;
    ASSERT(pUrb);

    // Store BulkOut PipeId
    pTxContext->BulkOutPipeId = BulkOutPipeId;

    //Initialize a tx bulk urb
    RTusb_fill_bulk_urb(pUrb, pAd->pUsb_Dev,
                        usb_sndbulkpipe(pAd->pUsb_Dev, 1),
                        (PUCHAR) &pTxContext->TransferBuffer->TxDesc,
                        pTxContext->BulkOutSize, Func, pTxContext);
}

static VOID RTUSBInitRxDesc(IN PRTMP_ADAPTER pAd, IN PRX_CONTEXT pRxContext)
{
    PURB pUrb = pRxContext->pUrb;
    ASSERT(pUrb);

    //Initialize a rx bulk urb
    RTusb_fill_bulk_urb(pUrb,
                        pAd->pUsb_Dev,
                        usb_rcvbulkpipe(pAd->pUsb_Dev, 1),
                        pRxContext->TransferBuffer,
                        BUFFER_SIZE, RTUSBBulkRxComplete, pRxContext);
}

static VOID RTUSBBulkOutDataPacket(IN PRTMP_ADAPTER pAd,
                                   IN UCHAR BulkOutPipeId, IN UCHAR Index)
{
    PTX_CONTEXT pTxContext;
    PURB pUrb;
    int ret = 0;
    unsigned long IrqFlags;

    NdisAcquireSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
    if (pAd->BulkOutPending[BulkOutPipeId] == TRUE) {
        NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        DBUG(9, "returning; already locked\n");
        return;
    }
    pAd->BulkOutPending[BulkOutPipeId] = TRUE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);

    pTxContext = &(pAd->TxContext[BulkOutPipeId][Index]);

    // Increase Total transmit byte counter
    pAd->RalinkCounters.TransmittedByteCount += pTxContext->BulkOutSize;

    // Clear Data flag
    RTUSB_CLEAR_BULK_FLAG(pAd,(fRTUSB_BULK_OUT_DATA_FRAG << BulkOutPipeId));
    RTUSB_CLEAR_BULK_FLAG(pAd,(fRTUSB_BULK_OUT_DATA_NORMAL << BulkOutPipeId));

    if (pTxContext->bWaitingBulkOut != TRUE) {
        DBUG(9,"bWaitingBulkOut != TRUE, Index %d, NextBulkOutIndex %d\n",
             Index, pAd->NextBulkOutIndex[BulkOutPipeId]);
        NdisAcquireSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        pAd->BulkOutPending[BulkOutPipeId] = FALSE;
        NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        return;
    } else if (pTxContext->BulkOutSize == 0) {
        //
        // This may happen on CCX Leap Ckip or Cmic
        // When the Key was been set not on time.
        // We will break it when the Key was Zero on RTUSBHardTransmit
        // And this will cause deadlock that the TxContext always InUse.
        //
        NdisAcquireSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);

        pTxContext->InUse = FALSE;
        pTxContext->LastOne = FALSE;
        pTxContext->IRPPending = FALSE;
        pTxContext->bWaitingBulkOut = FALSE;
        pTxContext->BulkOutSize = 0;
        pAd->NextBulkOutIndex[BulkOutPipeId] =
            (pAd->NextBulkOutIndex[BulkOutPipeId] + 1) % TX_RING_SIZE;
        pAd->BulkOutPending[BulkOutPipeId] = FALSE;
        NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        DBUG(9, "failed; pTxContext->BulkOutSize==0\n");
        return;
    #if 0
    } else if (!OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED)
               && !(pAd->PortCfg.BssType == BSS_MONITOR &&
                    pAd->bAcceptRFMONTx == TRUE)) {
    #else
    } else if (!OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED)) {
    #endif
        //
        // Since there is no connection, so we need to empty the Tx Bulk out Ring.
        //
        while (atomic_read(&pAd->TxCount) > 0) {
            OOPS("Failed, since NdisMediaStateDisconnected discard NextBulkOutIndex %d, NextIndex = %d\n",
                     pAd->NextBulkOutIndex[BulkOutPipeId],
                     pAd->NextTxIndex[BulkOutPipeId]);

            FREE_TX_RING(pAd, BulkOutPipeId, pTxContext);
            pAd->TxRingTotalNumber[BulkOutPipeId]--;    // sync. to TxCount
            pTxContext = &(pAd->TxContext[BulkOutPipeId][pAd->NextBulkOutIndex
                                                             [BulkOutPipeId]]);
        }

        NdisAcquireSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        pAd->BulkOutPending[BulkOutPipeId] = FALSE;
        NdisReleaseSpinLock(&pAd->BulkOutLock[BulkOutPipeId], IrqFlags);
        DBUG(9, "failed; no connection!\n");
        return;
    }

    // Init Tx context descriptor
    RTUSBInitTxDesc(pAd, pTxContext, BulkOutPipeId,
                    RTUSBBulkOutDataPacketComplete);

    pTxContext->IRPPending = TRUE;

    #ifdef RT73_DBG
    {
        extern char* dbug_type( int type, char* string );
        extern char* dbug_subtype( int type, int subtype, char* string );
        PHEADER_802_11 pHeader = (PHEADER_802_11) pTxContext->TransferBuffer->WirelessPacket;
        char type_string[32];
        char subtype_string[32];

        DBUG(4,"TX-%03x %s.%s A1=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A2=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A3=%02x:%02x:%02x:%02x:%02x:%02x [0x%x bytes]\n", 
             pHeader->Sequence,
             dbug_type(pHeader->FC.Type, type_string), 
             dbug_subtype(pHeader->FC.Type, pHeader->FC.SubType, subtype_string), 
             pHeader->Addr1[0], pHeader->Addr1[1], pHeader->Addr1[2],
             pHeader->Addr1[3], pHeader->Addr1[4], pHeader->Addr1[5],
             pHeader->Addr2[0], pHeader->Addr2[1], pHeader->Addr2[2],
             pHeader->Addr2[3], pHeader->Addr2[4], pHeader->Addr2[5],
             pHeader->Addr3[0], pHeader->Addr3[1], pHeader->Addr3[2],
             pHeader->Addr3[3], pHeader->Addr3[4], pHeader->Addr3[5],
             pTxContext->pUrb->actual_length);
        if (dbug_level >= 8)
            dbug_data( (PUCHAR)pHeader, pTxContext->pUrb->actual_length );
    }
    #endif

    pUrb = pTxContext->pUrb;
    if ((ret = rtusb_submit_urb(pUrb)) != 0) {
        ERR("Submit Tx URB failed %d\n", ret);
        return;
    }

    DBUG(9, "done.\n");
}

/*
	========================================================================

	Routine Description:

	Arguments:

	Return Value:

	Note: NULL frame use BulkOutPipeId = 0

	========================================================================
*/
static VOID RTUSBBulkOutNullFrame(IN PRTMP_ADAPTER pAd)
{
    PTX_CONTEXT pNullContext = &(pAd->NullContext);
    PURB pUrb;
    int ret = 0;
    unsigned long IrqFlags;

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    if (pAd->BulkOutPending[0] == TRUE) {
        NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);
        return;
    }
    pAd->BulkOutPending[0] = TRUE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    // Increase Total transmit byte counter
    pAd->RalinkCounters.TransmittedByteCount += pNullContext->BulkOutSize;

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutNullFrame \n");

    // Clear Null frame bulk flag
    RTUSB_CLEAR_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NULL);

    // Init Tx context descriptor
    RTUSBInitTxDesc(pAd, pNullContext, 0, RTUSBBulkOutNullFrameComplete);
    pNullContext->IRPPending = TRUE;

    #ifdef RT73_DBG
    {
        extern char* dbug_type( int type, char* string );
        extern char* dbug_subtype( int type, int subtype, char* string );
        PHEADER_802_11 pHeader = (PHEADER_802_11) pNullContext->TransferBuffer->WirelessPacket;
        char type_string[32];
        char subtype_string[32];

        DBUG(4,"TX-%03x %s.%s A1=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A2=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A3=%02x:%02x:%02x:%02x:%02x:%02x [0x%x bytes]\n", 
             pHeader->Sequence,
             dbug_type(pHeader->FC.Type, type_string), 
             dbug_subtype(pHeader->FC.Type, pHeader->FC.SubType, subtype_string), 
             pHeader->Addr1[0], pHeader->Addr1[1], pHeader->Addr1[2],
             pHeader->Addr1[3], pHeader->Addr1[4], pHeader->Addr1[5],
             pHeader->Addr2[0], pHeader->Addr2[1], pHeader->Addr2[2],
             pHeader->Addr2[3], pHeader->Addr2[4], pHeader->Addr2[5],
             pHeader->Addr3[0], pHeader->Addr3[1], pHeader->Addr3[2],
             pHeader->Addr3[3], pHeader->Addr3[4], pHeader->Addr3[5],
             pNullContext->pUrb->actual_length);
    }
    #endif

    pUrb = pNullContext->pUrb;
    if ((ret = rtusb_submit_urb(pUrb)) != 0) {
        ERR("Submit Tx URB failed %d\n", ret);
        return;
    }

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutNullFrame \n");
}

/*
	========================================================================

	Routine Description:

	Arguments:

	Return Value:

	Note: MLME use BulkOutPipeId = 0

	========================================================================
*/
static VOID RTUSBBulkOutMLMEPacket(IN PRTMP_ADAPTER pAd, IN UCHAR Index)
{
    PTX_CONTEXT pMLMEContext;
    PURB pUrb;
    int ret = 0;
    unsigned long IrqFlags;

    pMLMEContext = &pAd->MLMEContext[Index];

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    if (pAd->BulkOutPending[0] == TRUE) {
        NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);
        return;
    }
    pAd->BulkOutPending[0] = TRUE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    // Increase Total transmit byte counter
    pAd->RalinkCounters.TransmittedByteCount += pMLMEContext->BulkOutSize;

    // Clear MLME bulk flag
    RTUSB_CLEAR_BULK_FLAG(pAd, fRTUSB_BULK_OUT_MLME);

    DBGPRINT_RAW(RT_DEBUG_INFO,
                 "RTUSBBulkOutMLMEPacket::PrioRingFirstIndex = %d, PrioRingTxCnt = %d, PopMgmtIndex = %d, PushMgmtIndex = %d, NextMLMEIndex = %d\n",
                 pAd->PrioRingFirstIndex, pAd->PrioRingTxCnt,
                 pAd->PopMgmtIndex, pAd->PushMgmtIndex,
                 pAd->NextMLMEIndex);

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutMLMEPacket\n");

    // Init Tx context descriptor
    RTUSBInitTxDesc(pAd, pMLMEContext, 0, RTUSBBulkOutMLMEPacketComplete);
    pMLMEContext->IRPPending = TRUE;

    #ifdef RT73_DBG
    {
        extern char* dbug_type( int type, char* string );
        extern char* dbug_subtype( int type, int subtype, char* string );
        PHEADER_802_11 pHeader = (PHEADER_802_11) pMLMEContext->TransferBuffer->WirelessPacket;
        char type_string[32];
        char subtype_string[32];

        DBUG(4,"TX-%03x %s.%s A1=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A2=%02x:%02x:%02x:%02x:%02x:%02x "
                             "A3=%02x:%02x:%02x:%02x:%02x:%02x [0x%x bytes]\n", 
             pHeader->Sequence,
             dbug_type(pHeader->FC.Type, type_string), 
             dbug_subtype(pHeader->FC.Type, pHeader->FC.SubType, subtype_string), 
             pHeader->Addr1[0], pHeader->Addr1[1], pHeader->Addr1[2],
             pHeader->Addr1[3], pHeader->Addr1[4], pHeader->Addr1[5],
             pHeader->Addr2[0], pHeader->Addr2[1], pHeader->Addr2[2],
             pHeader->Addr2[3], pHeader->Addr2[4], pHeader->Addr2[5],
             pHeader->Addr3[0], pHeader->Addr3[1], pHeader->Addr3[2],
             pHeader->Addr3[3], pHeader->Addr3[4], pHeader->Addr3[5],
             pMLMEContext->pUrb->actual_length);
        if (dbug_level >= 8)
            dbug_data( (PUCHAR)pHeader, pMLMEContext->pUrb->actual_length );
    }
    #endif

    pUrb = pMLMEContext->pUrb;
    if ((ret = rtusb_submit_urb(pUrb)) != 0) {
        ERR("Submit MLME URB failed %d\n", ret);
        return;
    }

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutMLMEPacket \n");
}

#if 0
/*
	========================================================================

	Routine Description:

	Arguments:

	Return Value:

	Note: PsPoll use BulkOutPipeId = 0

	========================================================================
*/
static VOID RTUSBBulkOutPsPoll(IN PRTMP_ADAPTER pAd)
{
    PTX_CONTEXT pPsPollContext = &(pAd->PsPollContext);
    PURB pUrb;
    int ret = 0;
    unsigned long IrqFlags;

    NdisAcquireSpinLock(&pAd->BulkOutLock[0], IrqFlags);
    if (pAd->BulkOutPending[0] == TRUE) {
        NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);
        return;
    }
    pAd->BulkOutPending[0] = TRUE;
    NdisReleaseSpinLock(&pAd->BulkOutLock[0], IrqFlags);

    DBGPRINT_RAW(RT_DEBUG_INFO, "--->RTUSBBulkOutPsPoll \n");

    // Clear PS-Poll bulk flag
    RTUSB_CLEAR_BULK_FLAG(pAd, fRTUSB_BULK_OUT_PSPOLL);

    // Init Tx context descriptor
    RTUSBInitTxDesc(pAd, pPsPollContext, 0, RTUSBBulkOutPsPollComplete);
    pPsPollContext->IRPPending = TRUE;

    pUrb = pPsPollContext->pUrb;
    if ((ret = rtusb_submit_urb(pUrb)) != 0) {
        ERR("Submit Tx URB failed %d\n", ret);
        return;
    }

    DBGPRINT_RAW(RT_DEBUG_INFO, "<---RTUSBBulkOutPsPoll \n");
}
#endif

/*
	========================================================================

	Routine Description:
	USB_RxPacket initializes a URB and uses the Rx IRP to submit it
	to USB. It checks if an Rx Descriptor is available and passes the
	the coresponding buffer to be filled. If no descriptor is available
	fails the request. When setting the completion routine we pass our
	Adapter Object as Context.

	Arguments:

	Return Value:
		TRUE			found matched tuple cache
		FALSE			no matched found

	Note:

	========================================================================
*/
VOID RTUSBBulkReceive(IN PRTMP_ADAPTER pAd)
{
    PRX_CONTEXT pRxContext;
    int ret;

    if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST |
                            fRTMP_ADAPTER_RADIO_OFF     |
                            fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS)) {
        DBUG(4,"SORRY, NIC_NOT_EXIST=%d, RADIO_OFF=%d, RESET_PIPE=%d\n",
                 RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST),
                 RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF),
                 RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS));
        return;
    }

    pRxContext = &(pAd->RxContext[pAd->NextRxBulkInIndex]);
    pRxContext->InUse = TRUE;
    pAd->NextRxBulkInIndex = (pAd->NextRxBulkInIndex + 1) % RX_RING_SIZE;

    atomic_set(&pRxContext->IrpLock, IRPLOCK_CANCELABLE);
    atomic_inc(&pAd->PendingRx);

    // Init Rx context descriptor
    #if 0
    // GRH 1/11/2008: why would they do this?  to make it slower?
    memset(pRxContext->TransferBuffer, 0, BUFFER_SIZE);
    #else
    memset(pRxContext->TransferBuffer, 0, sizeof(RXD_STRUC));
    #endif
    RTUSBInitRxDesc(pAd, pRxContext);

    DBUG( 4, "usb_submit_urb(urb=%p,buf=%p)\n", 
              pRxContext->pUrb, pRxContext->TransferBuffer );
    if ((ret = rtusb_submit_urb(pRxContext->pUrb)) != 0) {
        ERR("usb_submit_urb(%p) failed, error %d\n", pRxContext->pUrb, ret);
    }
}

/*
	========================================================================

	Routine Description:
	RTUSBBulkReceive cannot be called when in an interrupt as it raises a
	bug on ARM, so we schedule this function so that it calls
	RTUSBBulkReceive outside of interrupt context.

	Arguments: ulong data, memory address to pAd structure.

	Return Value:

	Note:

	========================================================================
*/

VOID rtusb_bulkrx(IN unsigned long data)
{
    RTUSBBulkReceive((PRTMP_ADAPTER) data);
}

VOID RTUSBBulkRxHandle(IN unsigned long data)
{
	purbb_t 		pUrb       = (purbb_t)data;
	PRX_CONTEXT 	pRxContext = (PRX_CONTEXT)pUrb->context;
	
	/* device had been closed */
	if (!RTMP_TEST_FLAG((PRTMP_ADAPTER)pRxContext->pAd, fRTMP_ADAPTER_REMOVE_IN_PROGRESS))
        RTUSBRxPacket(data);
}

VOID RTUSBKickBulkOut(IN PRTMP_ADAPTER pAd)
{
    DBUG_HERE(9);

    if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST          |
                             fRTMP_ADAPTER_HALT_IN_PROGRESS       |
                             fRTMP_ADAPTER_RESET_IN_PROGRESS      |
                             fRTMP_ADAPTER_RESET_PIPE_IN_PROGRESS |
                             fRTMP_ADAPTER_RADIO_OFF)) {
        // 1. Data Fragment has highest priority
        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_FRAG)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 0)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 0, pAd->NextBulkOutIndex[0]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_FRAG_2)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 1)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 1, pAd->NextBulkOutIndex[1]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_FRAG_3)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 2)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 2, pAd->NextBulkOutIndex[2]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_FRAG_4)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 3)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 3, pAd->NextBulkOutIndex[3]);
        }
        #if 0
        // 2. PS-Poll frame is next
        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_PSPOLL))
            RTUSBBulkOutPsPoll(pAd);
        else 
        #endif
        // 5. Mlme frame is next
        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_MLME))
            RTUSBBulkOutMLMEPacket(pAd, pAd->PrioRingFirstIndex);
        // 6. Data frame normal is next
        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NORMAL)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 0)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 0, pAd->NextBulkOutIndex[0]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NORMAL_2)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 1)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 1, pAd->NextBulkOutIndex[1]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NORMAL_3)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 2)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 2, pAd->NextBulkOutIndex[2]);
        }

        if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NORMAL_4)) {
            if ((!LOCAL_TX_RING_EMPTY(pAd, 3)) &&
                (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
                RTUSBBulkOutDataPacket(pAd, 3, pAd->NextBulkOutIndex[3]);
        }
        // 7. Null frame is the last
        else if (RTUSB_TEST_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NULL)) {
            if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS))
                RTUSBBulkOutNullFrame(pAd);
        }
        // 8. No data avaliable
        else {
            DBUG(9,"No data to kick!\n");
        }
    }

    DBUG(9, "done.\n");
}

VOID RTUSBCleanUpDataBulkOutQueue(IN PRTMP_ADAPTER pAd)
{
    UCHAR Idx;

    DBGPRINT(RT_DEBUG_TRACE, "--->CleanUpDataBulkOutQueue\n");

    for (Idx = 0; Idx < 4; Idx++) {
        PTX_CONTEXT pTxContext;
        unsigned long IrqFlags;

        while (!LOCAL_TX_RING_EMPTY(pAd, Idx)) {
            pTxContext = &(pAd->TxContext[Idx][pAd->NextBulkOutIndex[Idx]]);
            pTxContext->LastOne = FALSE;
            pTxContext->InUse = FALSE;
            pTxContext->bWaitingBulkOut = FALSE;
            pAd->NextBulkOutIndex[Idx] =
                (pAd->NextBulkOutIndex[Idx] + 1) % TX_RING_SIZE;
        }
        NdisAcquireSpinLock(&pAd->BulkOutLock[Idx], IrqFlags);
        pAd->BulkOutPending[Idx] = FALSE;
        NdisReleaseSpinLock(&pAd->BulkOutLock[Idx], IrqFlags);
    }

    DBGPRINT(RT_DEBUG_TRACE, "<---CleanUpDataBulkOutQueue\n");
}

VOID RTUSBCleanUpMLMEBulkOutQueue(IN PRTMP_ADAPTER pAd)
{
    unsigned long IrqFlags;

    DBGPRINT(RT_DEBUG_TRACE, "--->CleanUpMLMEBulkOutQueue\n");

    NdisAcquireSpinLock(&pAd->MLMEQLock, IrqFlags);
    while (pAd->PrioRingTxCnt > 0) {
        pAd->MLMEContext[pAd->PrioRingFirstIndex].InUse = FALSE;

        pAd->PrioRingFirstIndex++;
        if (pAd->PrioRingFirstIndex >= PRIO_RING_SIZE) {
            pAd->PrioRingFirstIndex = 0;
        }

        pAd->PrioRingTxCnt--;
    }
    NdisReleaseSpinLock(&pAd->MLMEQLock, IrqFlags);

    DBGPRINT(RT_DEBUG_TRACE, "<---CleanUpMLMEBulkOutQueue\n");
}

static VOID RTUSBCancelPendingBulkInIRP(IN PRTMP_ADAPTER pAd)
{
    UINT i;

    DBGPRINT_RAW(RT_DEBUG_TRACE, "--->RTUSBCancelPendingBulkInIRP\n");
    for (i = 0; i < RX_RING_SIZE; i++) {
        PRX_CONTEXT pRxContext = &(pAd->RxContext[i]);
        if (atomic_read(&pRxContext->IrpLock) == IRPLOCK_CANCELABLE) {
            DBGPRINT(RT_DEBUG_TRACE, "usb_kill_urb(%p)\n", pRxContext->pUrb);
            RTUSB_UNLINK_URB(pRxContext->pUrb);
        } else
            DBGPRINT( RT_DEBUG_TRACE, "CANNOT usb_kill_urb(%p), naughty IrpLock\n",
                      pRxContext->pUrb );
        atomic_set(&pRxContext->IrpLock, IRPLOCK_CANCE_START);
    }
    DBGPRINT_RAW(RT_DEBUG_TRACE, "<---RTUSBCancelPendingBulkInIRP\n");
}

VOID RTUSBCancelPendingBulkOutIRP(IN PRTMP_ADAPTER pAd)
{
    PTX_CONTEXT pTxContext;
    PTX_CONTEXT pMLMEContext;
    #if 0
    PTX_CONTEXT pBeaconContext;
    #endif
    PTX_CONTEXT pNullContext;
    PTX_CONTEXT pPsPollContext;
    PTX_CONTEXT pRTSContext;
    UINT i, Idx;
    unsigned long IrqFlags;

    for (Idx = 0; Idx < 4; Idx++) {
        for (i = 0; i < TX_RING_SIZE; i++) {
            pTxContext = &(pAd->TxContext[Idx][i]);

            if (pTxContext->IRPPending == TRUE) {

                // Get the USB_CONTEXT and cancel it's IRP; the completion routine will itself
                // remove it from the HeadPendingSendList and NULL out HeadPendingSendList
                //      when the last IRP on the list has been  cancelled; that's how we exit this loop
                //

                RTUSB_UNLINK_URB(pTxContext->pUrb);

                // Sleep 200 microseconds to give cancellation time to work
                RTMPusecDelay(200);
            }
        }
    }

    for (i = 0; i < PRIO_RING_SIZE; i++) {
        pMLMEContext = &(pAd->MLMEContext[i]);

        if (pMLMEContext->IRPPending == TRUE) {

            // Get the USB_CONTEXT and cancel it's IRP; the completion routine will itself
            // remove it from the HeadPendingSendList and NULL out HeadPendingSendList
            //      when the last IRP on the list has been  cancelled; that's how we exit this loop
            //

            RTUSB_UNLINK_URB(pMLMEContext->pUrb);

            // Sleep 200 microsecs to give cancellation time to work
            RTMPusecDelay(200);
        }
    }

#if 0
    for (i = 0; i < BEACON_RING_SIZE; i++) {
        pBeaconContext = &(pAd->BeaconContext[i]);

        if (pBeaconContext->IRPPending == TRUE) {

            // Get the USB_CONTEXT and cancel it's IRP; the completion routine will itself
            // remove it from the HeadPendingSendList and NULL out HeadPendingSendList
            //      when the last IRP on the list has been  cancelled; that's how we exit this loop
            //

            RTUSB_UNLINK_URB(pBeaconContext->pUrb);

            // Sleep 200 microsecs to give cancellation time to work
            RTMPusecDelay(200);
        }
    }
#endif

    pNullContext = &(pAd->NullContext);
    if (pNullContext->IRPPending == TRUE)
        RTUSB_UNLINK_URB(pNullContext->pUrb);

    pRTSContext = &(pAd->RTSContext);
    if (pRTSContext->IRPPending == TRUE)
        RTUSB_UNLINK_URB(pRTSContext->pUrb);

    pPsPollContext = &(pAd->PsPollContext);
    if (pPsPollContext->IRPPending == TRUE)
        RTUSB_UNLINK_URB(pPsPollContext->pUrb);

    for (Idx = 0; Idx < 4; Idx++) {
        NdisAcquireSpinLock(&pAd->BulkOutLock[Idx], IrqFlags);
        pAd->BulkOutPending[Idx] = FALSE;
        NdisReleaseSpinLock(&pAd->BulkOutLock[Idx], IrqFlags);
    }
}

VOID RTUSBCancelPendingIRPs(IN PRTMP_ADAPTER pAd)
{
    RTUSBCancelPendingBulkInIRP(pAd);
    RTUSBCancelPendingBulkOutIRP(pAd);
}
